import json
import logging
import requests
import threading

from segmind.entities import Experiment, Metric, Run, RunInfo, ViewType
from segmind.exceptions import MlflowException
from segmind.protos import errorcodes_pb2
from segmind.protos.service_lite_pb2 import (CreateExperiment, CreateRun,
                                             DeleteExperiment, DeleteRun,
                                             DeleteTag, GetExperiment,
                                             GetExperimentByName,
                                             GetMetricHistory, GetRun,
                                             ListExperiments, LogArtifact,
                                             LogBatch, LogMetric, LogModel,
                                             LogParam, MlflowService,
                                             RestoreExperiment, RestoreRun,
                                             RunTag, SearchRuns,
                                             SetExperimentTag, SetTag,
                                             UpdateExperiment, UpdateRun)
from segmind.store.tracking.abstract_store import AbstractStore
from segmind.utils.proto_json_utils import message_to_json
from segmind.utils.rest_utils import (call_endpoint,
                                      extract_api_info_for_service)

_PATH_PREFIX = '/api/2.0'
_METHOD_TO_INFO = extract_api_info_for_service(MlflowService, _PATH_PREFIX)
_logger = logging.getLogger(__name__)


def artifact_uploader(artifact, response_proto):
    # POST artifact to S3 presigned URL
    # FIXME: Replace _ with - in fields key names in MLflow-lite
    data = json.loads(message_to_json(response_proto.fields))
    data = {k.replace('_', '-'): v for k, v in data.items()}

    with open(artifact.path, 'rb') as f:
        files = {'file': (artifact.path, f)}
        s3_response = requests.post(response_proto.url, data=data, files=files)
    _logger.info(
        f'Artifact upload HTTP status code: {s3_response.status_code}')


class RestStore(AbstractStore):
    """Client for a remote tracking server accessed via REST API calls."""

    def __init__(self, ):
        super(RestStore, self).__init__()

    def _call_endpoint(self, api, json_body):
        endpoint, method = _METHOD_TO_INFO[api]
        response_proto = api.Response()

        return call_endpoint(endpoint, method, json_body, response_proto)

    def list_experiments(self, view_type=ViewType.ACTIVE_ONLY):
        """
        :return: a list of all known Experiment objects
        """
        req_body = message_to_json(ListExperiments(view_type=view_type))
        response_proto = self._call_endpoint(ListExperiments, req_body)
        return [
            Experiment.from_proto(experiment_proto)
            for experiment_proto in response_proto.experiments
        ]

    def create_experiment(self, name, artifact_location=None):
        """Create a new experiment. If an experiment with the given name
        already exists, throws exception.

        :param name: Desired name for an experiment

        :return: experiment_id (string) for the newly created experiment
                 if successful, else None
        """
        req_body = message_to_json(
            CreateExperiment(name=name, artifact_location=artifact_location))
        response_proto = self._call_endpoint(CreateExperiment, req_body)
        return response_proto.experiment_id

    def get_experiment(self, experiment_id):
        """Fetch the experiment from the backend store.

        :param experiment_id: String id for the experiment

        :return: A single :py:class:`segmind_track.entities.Experiment`
                 object if it exists, otherwise raises an Exception.
        """
        req_body = message_to_json(
            GetExperiment(experiment_id=str(experiment_id)))
        response_proto = self._call_endpoint(GetExperiment, req_body)
        return Experiment.from_proto(response_proto.experiment)

    def delete_experiment(self, experiment_id):
        req_body = message_to_json(
            DeleteExperiment(experiment_id=str(experiment_id)))
        self._call_endpoint(DeleteExperiment, req_body)

    def restore_experiment(self, experiment_id):
        req_body = message_to_json(
            RestoreExperiment(experiment_id=str(experiment_id)))
        self._call_endpoint(RestoreExperiment, req_body)

    def rename_experiment(self, experiment_id, new_name):
        req_body = message_to_json(
            UpdateExperiment(
                experiment_id=str(experiment_id), new_name=new_name))
        self._call_endpoint(UpdateExperiment, req_body)

    def get_run(self, run_id):
        """Fetch the run from backend store.

        :param run_id: Unique identifier for the run

        :return: A single Run object if it exists,
                 otherwise raises an Exception
        """
        req_body = message_to_json(GetRun(run_uuid=run_id, run_id=run_id))
        response_proto = self._call_endpoint(GetRun, req_body)
        return Run.from_proto(response_proto.run)

    def update_run_info(self, run_id, run_status, end_time):
        """Updates the metadata of the specified run."""
        req_body = message_to_json(
            UpdateRun(
                run_uuid=run_id,
                run_id=run_id,
                status=run_status,
                end_time=end_time))
        response_proto = self._call_endpoint(UpdateRun, req_body)
        return RunInfo.from_proto(response_proto.run_info)

    def create_run(self, experiment_id, start_time, tags):
        """Create a run under the specified experiment ID, setting the run's
        status to "RUNNING" and the start time to the current time.

        :param experiment_id: ID of the experiment for this run
        :param user_id: ID of the user launching this run
        :param source_type: Enum (integer) describing the source of the run

        :return: The created Run object
        """
        tag_protos = [
            RunTag(key=tagkey, value=tagvalue)
            for tagkey, tagvalue in tags.items()
        ]
        req_body = message_to_json(
            CreateRun(
                experiment_id=str(experiment_id),
                start_time=start_time,
                tags=tag_protos))
        response_proto = self._call_endpoint(CreateRun, req_body)
        run = Run.from_proto(response_proto.run)
        return run

    def log_metric(self, run_id, metric):
        """Log a metric for the specified run.

        :param run_id: String id for the run
        :param metric: Metric instance to log
        """
        req_body = message_to_json(
            LogMetric(
                run_uuid=run_id,
                run_id=run_id,
                key=metric.key,
                value=metric.value,
                timestamp=metric.timestamp,
                step=metric.step,
                epoch=metric.epoch,
                tags=metric.tags))
        self._call_endpoint(LogMetric, req_body)

    def log_param(self, run_id, param):
        """Log a param for the specified run.

        :param run_id: String id for the run
        :param param: Param instance to log
        """
        req_body = message_to_json(
            LogParam(
                run_uuid=run_id,
                run_id=run_id,
                key=param.key,
                value=param.value,
                tags=param.tags))
        self._call_endpoint(LogParam, req_body)

    def log_artifact(self, run_id, experiment_id, artifact):
        """Log an artifact for the specified run.

        :param run_id: String ID for the run
        :param experiment_id: String ID for the experiment
        :param artifact: Artifact instance to log
        """
        # Generate a presigned S3 POST URL
        req_body = message_to_json(
            LogArtifact(
                run_id=run_id,
                experiment_id=experiment_id,
                key=artifact.key,
                type=artifact.artifact_type,
                timestamp=artifact.timestamp,
                size=artifact.size,
                prediction=artifact.prediction,
                ground_truth=artifact.ground_truth,
                step=artifact.step))
        response_proto = self._call_endpoint(LogArtifact, req_body)
        if response_proto is None:
            _logger.warning(
                f'Artifact log HTTP status code: {response_proto.status_code}')

        # # POST artifact to S3 presigned URL
        # # FIXME: Replace _ with - in fields key names in MLflow-lite
        # data = json.loads(message_to_json(response_proto.fields))
        # data = {k.replace('_','-'): v for k,v in data.items()}

        # with open(artifact.path, 'rb') as f:
        #     files = {'file': (artifact.path, f)}
        #     s3_response = requests.post(
        #         response_proto.url,
        #         data=data, files=files
        #     )
        # _logger.info(f"Artifact upload HTTP status code: {s3_response.status_code}")  # noqa
        # artifact_uploader(artifact, response_proto)

        upload_thread = threading.Thread(
            target=artifact_uploader,
            kwargs={
                'artifact': artifact,
                'response_proto': response_proto
            },
        )
        upload_thread.daemon = False
        # print('starting upload in a thread ...')
        upload_thread.start()
        # print('thread started ...')

    def set_project_tag(self, experiment_id, tag):
        """Set a tag for the specified experiment.

        :param experiment_id: String ID of the experiment
        :param tag: ExperimentRunTag instance to log
        """
        req_body = message_to_json(
            SetExperimentTag(
                experiment_id=experiment_id, key=tag.key, value=tag.value))
        self._call_endpoint(SetExperimentTag, req_body)

    def set_tag(self, run_id, tag):
        """Set a tag for the specified run.

        :param run_id: String ID of the run
        :param tag: RunTag instance to log
        """
        req_body = message_to_json(
            SetTag(
                run_uuid=run_id, run_id=run_id, key=tag.key, value=tag.value))
        self._call_endpoint(SetTag, req_body)

    def delete_tag(self, run_id, key):
        """Delete a tag from a run. This is irreversible.

        :param run_id: String ID of the run
        :param key: Name of the tag
        """
        req_body = message_to_json(DeleteTag(run_id=run_id, key=key))
        self._call_endpoint(DeleteTag, req_body)

    def get_metric_history(self, run_id, metric_key):
        """Return all logged values for a given metric.

        :param run_id: Unique identifier for run
        :param metric_key: Metric name within the run

        :return: A list of :py:class:`segmind_track.entities.Metric`
                 entities if logged, else empty list
        """
        req_body = message_to_json(
            GetMetricHistory(
                run_uuid=run_id, run_id=run_id, metric_key=metric_key))
        response_proto = self._call_endpoint(GetMetricHistory, req_body)
        return [Metric.from_proto(metric) for metric in response_proto.metrics]

    def _search_runs(self, experiment_ids, filter_string, run_view_type,
                     max_results, order_by, page_token):
        experiment_ids = [
            str(experiment_id) for experiment_id in experiment_ids
        ]
        sr = SearchRuns(
            experiment_ids=experiment_ids,
            filter=filter_string,
            run_view_type=ViewType.to_proto(run_view_type),
            max_results=max_results,
            order_by=order_by,
            page_token=page_token)
        req_body = message_to_json(sr)
        response_proto = self._call_endpoint(SearchRuns, req_body)
        runs = [Run.from_proto(proto_run) for proto_run in response_proto.runs]

        next_page_token = None
        if response_proto.next_page_token:
            next_page_token = response_proto.next_page_token
        return runs, next_page_token

    def delete_run(self, run_id):
        req_body = message_to_json(DeleteRun(run_id=run_id))
        self._call_endpoint(DeleteRun, req_body)

    def restore_run(self, run_id):
        req_body = message_to_json(RestoreRun(run_id=run_id))
        self._call_endpoint(RestoreRun, req_body)

    def get_experiment_by_name(self, experiment_name):
        try:
            req_body = message_to_json(
                GetExperimentByName(experiment_name=experiment_name))
            response_proto = self._call_endpoint(GetExperimentByName, req_body)
            return Experiment.from_proto(response_proto.experiment)
        except MlflowException as e:
            if e.error_code == errorcodes_pb2.ErrorCode.Name(
                    errorcodes_pb2.RESOURCE_DOES_NOT_EXIST):
                return None
            elif e.error_code == errorcodes_pb2.ErrorCode.Name(
                    errorcodes_pb2.REQUEST_LIMIT_EXCEEDED):
                raise e
            # Fall back to using ListExperiments-based implementation.
            for experiment in self.list_experiments(ViewType.ALL):
                if experiment.name == experiment_name:
                    return experiment
            return None

    def log_batch(self, run_id, metrics, params, tags):
        metric_protos = [metric.to_proto() for metric in metrics]
        param_protos = [param.to_proto() for param in params]
        # tag_protos = [tag.to_proto() for tag in tags]
        tag_protos = tags
        req_body = message_to_json(
            LogBatch(
                metrics=metric_protos,
                params=param_protos,
                tags=tag_protos,
                run_id=run_id))
        self._call_endpoint(LogBatch, req_body)

    def record_logged_model(self, run_id, mlflow_model):
        req_body = message_to_json(
            LogModel(run_id=run_id, model_json=mlflow_model.to_json()))
        self._call_endpoint(LogModel, req_body)

    def set_terminated(self, run_id, status, end_time):
        self.update_run_info(
            run_id=run_id, run_status=status, end_time=end_time)


class DatabricksRestStore(RestStore):
    """Databricks-specific RestStore implementation that provides different
    fallback.

    behavior when hitting the GetExperimentByName REST API fails
    - in particular, we only fall back to ListExperiments when
    the server responds with ENDPOINT_NOT_FOUND, rather than
    on all internal server errors. This implementation should
    be deprecated once GetExperimentByName is available everywhere.
    """

    def get_experiment_by_name(self, experiment_name):
        try:
            req_body = message_to_json(
                GetExperimentByName(experiment_name=experiment_name))
            response_proto = self._call_endpoint(GetExperimentByName, req_body)
            return Experiment.from_proto(response_proto.experiment)
        except MlflowException as e:
            if e.error_code == errorcodes_pb2.ErrorCode.Name(
                    errorcodes_pb2.RESOURCE_DOES_NOT_EXIST):
                return None
            elif e.error_code == errorcodes_pb2.ErrorCode.Name(
                    errorcodes_pb2.ENDPOINT_NOT_FOUND):
                # Fall back to using ListExperiments-based implementation.
                for experiment in self.list_experiments(ViewType.ALL):
                    if experiment.name == experiment_name:
                        return experiment
                return None
            raise e



from segmind.entities._mlflow_object import _MLflowObject
from segmind.protos.service_lite_pb2 import RunTag as ProtoRunTag


class RunTag(_MLflowObject):
    """Tag object associated with a run."""

    def __init__(self, key, value):
        self._key = key
        self._value = value

    def __eq__(self, other):
        if type(other) is type(self):
            # TODO deep equality here?
            return self.__dict__ == other.__dict__
        return False

    @property
    def key(self):
        """String name of the tag."""
        return self._key

    @property
    def value(self):
        """String value of the tag."""
        return self._value

    def to_proto(self):
        param = ProtoRunTag()
        param.key = self.key
        param.value = self.value
        return param

    @classmethod
    def from_proto(cls, proto):
        return cls(proto.key, proto.value)



from __future__ import print_function

import os
import sys

from segmind.lite_extensions.client_utils import get_host_uri
from segmind.store.tracking import DEFAULT_LOCAL_FILE_AND_ARTIFACT_PATH
from segmind.store.tracking.rest_store import RestStore
from segmind.tracking._tracking_service.registry import TrackingStoreRegistry
from segmind.utils import env
from segmind.utils.file_utils import path_to_local_file_uri

_TRACKING_URI_ENV_VAR = 'MLFLOW_TRACKING_URI'

# Extra environment variables which take precedence for setting the
# basic/bearer
# auth on http requests.
_TRACKING_USERNAME_ENV_VAR = 'MLFLOW_TRACKING_USERNAME'
_TRACKING_PASSWORD_ENV_VAR = 'MLFLOW_TRACKING_PASSWORD'
_TRACKING_TOKEN_ENV_VAR = 'MLFLOW_TRACKING_TOKEN'
_TRACKING_INSECURE_TLS_ENV_VAR = 'MLFLOW_TRACKING_INSECURE_TLS'

# TO DO:
# Depreciate this
_tracking_uri = get_host_uri()


def is_tracking_uri_set():
    """Returns True if the tracking URI has been set, False otherwise."""
    if _tracking_uri or env.get_env(_TRACKING_URI_ENV_VAR):
        return True
    return False


# Remove this
def set_tracking_uri(uri):
    """Set the tracking server URI. This does not affect the currently active
    run (if one exists), but takes effect for successive runs.

    :param uri:

    - An empty string, or a local file path, prefixed with ``file:/``. Data
    is stored locally at the provided file (or ``./mlruns`` if empty).
    - An HTTP URI like ``https://my-tracking-server:5000``.
    - A Databricks workspace, provided as the string "databricks" or, to use a
    Databricks CLI
    `profile <https://github.com/databricks/databricks-cli#installation>`_,
    "databricks://<profileName>".
    """
    global _tracking_uri
    _tracking_uri = uri


# Remove this
def get_tracking_uri():
    """Get the current tracking URI. This may not correspond to the tracking
    URI of the currently active run, since the tracking URI can be updated via
    ``set_tracking_uri``.

    :return: The tracking URI.
    """
    global _tracking_uri
    if _tracking_uri is not None:
        return _tracking_uri
    elif env.get_env(_TRACKING_URI_ENV_VAR) is not None:
        return env.get_env(_TRACKING_URI_ENV_VAR)
    else:
        return path_to_local_file_uri(
            os.path.abspath(DEFAULT_LOCAL_FILE_AND_ARTIFACT_PATH))


def _get_rest_store(store_uri, **_):
    return RestStore()


_tracking_store_registry = TrackingStoreRegistry()

for scheme in ['http', 'https']:
    _tracking_store_registry.register(scheme, _get_rest_store)

_tracking_store_registry.register_entrypoints()


def _get_store(store_uri=None, artifact_uri=None):
    return _tracking_store_registry.get_store(store_uri, artifact_uri)


# TODO(sueann): move to a projects utils module
def _get_git_url_if_present(uri):
    """Return the path git_uri#sub_directory if the URI passed is a local path
    that's part of a Git repo, or returns the original URI otherwise.

    :param uri: The expanded uri
    :return: The git_uri#sub_directory if the uri is part of a Git repo,
             otherwise return the original uri
    """
    if '#' in uri:
        # Already a URI in git repo format
        return uri
    try:
        from git import (GitCommandNotFound, InvalidGitRepositoryError,
                         NoSuchPathError, Repo)
    except ImportError as e:
        print(
            'Notice: failed to import Git (the git executable is probably '
            'not on your PATH), so Git SHA is not available. Error: %s' % e,
            file=sys.stderr)
        return uri
    try:
        # Check whether this is part of a git repo
        repo = Repo(uri, search_parent_directories=True)

        # Repo url
        repo_url = 'file://%s' % repo.working_tree_dir

        # Sub directory
        rlpath = uri.replace(repo.working_tree_dir, '')
        if (rlpath == ''):
            git_path = repo_url
        elif (rlpath[0] == '/'):
            git_path = repo_url + '#' + rlpath[1:]
        else:
            git_path = repo_url + '#' + rlpath
        return git_path
    except (InvalidGitRepositoryError, GitCommandNotFound, ValueError,
            NoSuchPathError):
        return uri



import distutils.dir_util as dir_util
import os
import shutil

from segmind.store.artifact.artifact_repo import (ArtifactRepository,
                                                  verify_artifact_path)
from segmind.utils.file_utils import (get_file_info, list_all,
                                      local_file_uri_to_path, mkdir,
                                      relative_path_to_artifact_path)


class LocalArtifactRepository(ArtifactRepository):
    """Stores artifacts as files in a local directory."""

    def __init__(self, *args, **kwargs):
        super(LocalArtifactRepository, self).__init__(*args, **kwargs)
        self._artifact_dir = local_file_uri_to_path(self.artifact_uri)

    @property
    def artifact_dir(self):
        return self._artifact_dir

    def log_artifact(self, local_file, artifact_path=None):
        verify_artifact_path(artifact_path)
        # NOTE: The artifact_path is expected to be in posix format.
        # Posix paths work fine on windows but just in case we normalize it
        # here.
        if artifact_path:
            artifact_path = os.path.normpath(artifact_path)

        if artifact_path:
            artifact_dir = os.path.join(self.artifact_dir, artifact_path)
        else:
            artifact_dir = self.artifact_dir

        if not os.path.exists(artifact_dir):
            mkdir(artifact_dir)
        shutil.copy(local_file, artifact_dir)

    def _is_directory(self, artifact_path):
        # NOTE: The path is expected to be in posix format.
        # Posix paths work fine on windows but just in case we normalize it
        # here.
        path = os.path.normpath(artifact_path) if artifact_path else ''
        list_dir = os.path.join(self.artifact_dir,
                                path) if path else self.artifact_dir
        return os.path.isdir(list_dir)

    def log_artifacts(self, local_dir, artifact_path=None):
        verify_artifact_path(artifact_path)
        # NOTE: The artifact_path is expected to be in posix format.
        # Posix paths work fine on windows but just in case we normalize it
        # here.
        if artifact_path:
            artifact_path = os.path.normpath(artifact_path)

        if artifact_path:
            artifact_dir = os.path.join(self.artifact_dir, artifact_path)
        else:
            artifact_dir = self.artifact_dir

        if not os.path.exists(artifact_dir):
            mkdir(artifact_dir)
        dir_util.copy_tree(src=local_dir, dst=artifact_dir)

    def download_artifacts(self, artifact_path, dst_path=None):
        """Artifacts tracked by ``LocalArtifactRepository`` already exist on
        the local filesystem. If ``dst_path`` is ``None``, the absolute
        filesystem path of the specified artifact is returned. If ``dst_path``
        is not ``None``, the local artifact is copied to ``dst_path``.

        Args:
            artifact_path: Relative source path to the desired artifacts.
            dst_path: Absolute path of the local filesystem destination
                    directory to which to download the specified artifacts.
                    This directory must already exist. If unspecified, the
                    absolute path of the local artifact will be returned.

        :return: Absolute path of the local filesystem location containing the
                 desired artifacts.
        """
        if dst_path:
            return super(LocalArtifactRepository,
                         self).download_artifacts(artifact_path, dst_path)
        # NOTE: The artifact_path is expected to be in posix format.
        # Posix paths work fine on windows but just in case we normalize it
        # here.
        local_artifact_path = os.path.join(self.artifact_dir,
                                           os.path.normpath(artifact_path))
        if not os.path.exists(local_artifact_path):
            raise IOError('No such file or directory: \'{}\''.format(
                local_artifact_path))
        return os.path.abspath(local_artifact_path)

    def list_artifacts(self, path=None):
        # NOTE: The path is expected to be in posix format.
        # Posix paths work fine on windows but just in case we normalize it
        # here.
        if path:
            path = os.path.normpath(path)
        list_dir = os.path.join(self.artifact_dir,
                                path) if path else self.artifact_dir
        if os.path.isdir(list_dir):
            artifact_files = list_all(list_dir, full_path=True)
            infos = [
                get_file_info(
                    f,
                    relative_path_to_artifact_path(
                        os.path.relpath(f, self.artifact_dir)))
                for f in artifact_files
            ]
            return sorted(infos, key=lambda f: f.path)
        else:
            return []

    def _download_file(self, remote_file_path, local_path):
        # NOTE: The remote_file_path is expected to be in posix format.
        # Posix paths work fine on windows but just in case we normalize it
        # here.
        remote_file_path = os.path.join(self.artifact_dir,
                                        os.path.normpath(remote_file_path))
        shutil.copyfile(remote_file_path, local_path)



import os

from segmind.exceptions import MlflowliteException
from segmind.protos.errorcodes_pb2 import INVALID_PARAMETER_VALUE


def get_env(variable_name):
    if variable_name not in os.environ:
        raise MlflowliteException('{} not set'.format(variable_name),
                                  INVALID_PARAMETER_VALUE)
    return os.environ.get(variable_name)


def unset_variable(variable_name):
    if variable_name in os.environ:
        del os.environ[variable_name]



"""Utilities for validating user inputs such as metric names and parameter
names."""
import numbers
import os
import posixpath
import re

from segmind.exceptions import MlflowException
from segmind.lite_extensions.server_only.file_utils import _ROOT_DIR
from segmind.protos.errorcodes_pb2 import (INVALID_PARAMETER_VALUE,
                                           RESOURCE_DOES_NOT_EXIST)
# from segmind.store.db.db_types import DATABASE_ENGINES
from segmind.utils.string_utils import is_string_type

_VALID_PARAM_AND_METRIC_NAMES = re.compile(r'^[/\w.\- ]*$')

# Regex for valid run IDs: must be an alphanumeric string of length 1 to 256.
_RUN_ID_REGEX = re.compile(r'^[a-zA-Z0-9][\w\-]{0,255}$')

_EXPERIMENT_ID_REGEX = re.compile(r'^[a-zA-Z0-9][\w\-]{0,63}$')

_BAD_CHARACTERS_MESSAGE = (
    'Names may only contain alphanumerics, underscores (_), dashes (-), '
    'periods (.), spaces ( ), and slashes (/).')

MAX_PARAMS_TAGS_PER_BATCH = 100
MAX_METRICS_PER_BATCH = 1000
MAX_ENTITIES_PER_BATCH = 1000
MAX_BATCH_LOG_REQUEST_SIZE = int(1e6)
MAX_PARAM_VAL_LENGTH = 250
MAX_TAG_VAL_LENGTH = 5000
MAX_EXPERIMENT_TAG_KEY_LENGTH = 250
MAX_EXPERIMENT_TAG_VAL_LENGTH = 5000
MAX_ENTITY_KEY_LENGTH = 250

_UNSUPPORTED_DB_TYPE_MSG = 'Supported database engines are segmind_track rest'


def bad_path_message(name):
    return (
        'Names may be treated as files in certain cases, and must not resolve'
        "to other names when treated as such. This name would resolve to '%s'"
    ) % posixpath.normpath(name)


def path_not_unique(name):
    norm = posixpath.normpath(name)
    return norm != name or norm == '.' or norm.startswith(
        '..') or norm.startswith('/')


def _validate_metric_name(name):
    """Check that `name` is a valid metric name and raise an exception if it
    isn't."""
    if not _VALID_PARAM_AND_METRIC_NAMES.match(name):
        raise MlflowException(
            "Invalid metric name: '%s'. %s" % (name, _BAD_CHARACTERS_MESSAGE),
            INVALID_PARAMETER_VALUE)
    if path_not_unique(name):
        raise MlflowException(
            "Invalid metric name: '%s'. %s" % (name, bad_path_message(name)),
            INVALID_PARAMETER_VALUE)


def _validate_metric(key, value, timestamp, step):
    """Check that a param with the specified key, value, timestamp is valid and
    raise an exception if it isn't."""
    _validate_metric_name(key)
    if not isinstance(value, numbers.Number):
        raise MlflowException(
            "Got invalid value %s for metric '%s' (timestamp=%s). Please "
            'specify value as a valid double (64-bit floating point)'
            % (value, key, timestamp), INVALID_PARAMETER_VALUE)

    if not isinstance(timestamp, numbers.Number) or timestamp < 0:
        raise MlflowException(
            "Got invalid timestamp %s for metric '%s' (value=%s). Timestamp "
            'must be a nonnegative long (64-bit integer) '
            % (timestamp, key, value), INVALID_PARAMETER_VALUE)

    if not isinstance(step, numbers.Number):
        raise MlflowException(
            "Got invalid step %s for metric '%s' (value=%s). Step must be a "
            'valid long (64-bit integer).'
            % (step, key, value), INVALID_PARAMETER_VALUE)


def _validate_param(key, value):
    """Check that a param with the specified key & value is valid and raise an
    exception if it isn't."""
    _validate_param_name(key)
    _validate_length_limit('Param key', MAX_ENTITY_KEY_LENGTH, key)
    _validate_length_limit('Param value', MAX_PARAM_VAL_LENGTH, value)


def _validate_artifact(key, path):
    """Check that an artifact with the specified key & valid and raise an
    exception if it isn't."""
    _validate_artifact_name(key)
    _validate_artifact_path(path)


def _validate_tag(key, value):
    """Check that a tag with the specified key & value is valid and raise an
    exception if it isn't."""
    _validate_tag_name(key)
    _validate_length_limit('Tag key', MAX_ENTITY_KEY_LENGTH, key)
    _validate_length_limit('Tag value', MAX_TAG_VAL_LENGTH, value)


def _validate_experiment_tag(key, value):
    """Check that a tag with the specified key & value is valid and raise an
    exception if it isn't."""
    _validate_tag_name(key)
    _validate_length_limit('Tag key', MAX_EXPERIMENT_TAG_KEY_LENGTH, key)
    _validate_length_limit('Tag value', MAX_EXPERIMENT_TAG_VAL_LENGTH, value)


def _validate_param_name(name):
    """Check that `name` is a valid parameter name and raise an exception if it
    isn't."""
    if not _VALID_PARAM_AND_METRIC_NAMES.match(name):
        raise MlflowException(
            "Invalid parameter name: '%s'. %s" %
            (name, _BAD_CHARACTERS_MESSAGE), INVALID_PARAMETER_VALUE)
    if path_not_unique(name):
        raise MlflowException(
            "Invalid parameter name: '%s'. %s" %
            (name, bad_path_message(name)), INVALID_PARAMETER_VALUE)


def _validate_artifact_name(name):
    """Check that `name` is a valid artifact name and raise an exception if it
    isn't.

    # TODO: Warn the user if artifact with same `name` already exists in the
    current run.
    """


def _validate_artifact_path(path):
    """Check that `path` is a valid artifact local path and raise an exception
    if is isn't."""


def _validate_tag_name(name):
    """Check that `name` is a valid tag name and raise an exception if it
    isn't."""
    # Reuse param & metric check.
    if not _VALID_PARAM_AND_METRIC_NAMES.match(name):
        raise MlflowException(
            "Invalid tag name: '%s'. %s" % (name, _BAD_CHARACTERS_MESSAGE),
            INVALID_PARAMETER_VALUE)
    if path_not_unique(name):
        raise MlflowException(
            "Invalid tag name: '%s'. %s" % (name, bad_path_message(name)),
            INVALID_PARAMETER_VALUE)


def _validate_length_limit(entity_name, limit, value):
    if len(value) > limit:
        raise MlflowException(
            "%s '%s' had length %s, which exceeded length limit of %s" %
            (entity_name, value[:250], len(value), limit))


def _validate_run_id(run_id):
    """Check that `run_id` is a valid run ID and raise an exception if it
    isn't."""
    if _RUN_ID_REGEX.match(run_id) is None:
        raise MlflowException(
            "Invalid run ID: '%s'" % run_id,
            error_code=INVALID_PARAMETER_VALUE)


def _validate_run_id_exp_id(exp_id, run_id):
    """Check that `run_id` is a valid run ID and raise an exception if it
    isn't."""
    if _RUN_ID_REGEX.match(run_id) is None:
        raise MlflowException(
            "Invalid run ID: '%s'" % run_id,
            error_code=INVALID_PARAMETER_VALUE)

    if _EXPERIMENT_ID_REGEX.match(exp_id) is None:
        raise MlflowException(
            "Invalid exp ID: '%s'" % run_id,
            error_code=INVALID_PARAMETER_VALUE)

    if not os.path.isdir(os.path.join(_ROOT_DIR, exp_id, run_id)):
        raise MlflowException(
            'RunId: {} doesnot exist under expID: {}'.format(run_id, exp_id),
            error_code=RESOURCE_DOES_NOT_EXIST)


def _validate_experiment_id(exp_id):
    """Check that `experiment_id`is a valid string or None, raise an exception
    if it isn't."""
    if exp_id is not None and _EXPERIMENT_ID_REGEX.match(exp_id) is None:
        raise MlflowException(
            "Invalid experiment ID: '%s'" % exp_id,
            error_code=INVALID_PARAMETER_VALUE)


def _validate_batch_limit(entity_name, limit, length):
    if length > limit:
        error_msg = (
            'A batch logging request can contain at most {limit} {name}. '
            'Got {count} {name}. Please split up {name} across multiple '
            'requests and try again.').format(
                name=entity_name, count=length, limit=limit)
        raise MlflowException(error_msg, error_code=INVALID_PARAMETER_VALUE)


def _validate_batch_log_limits(metrics, params, tags):
    """Validate that the provided batched logging arguments are within expected
    limits."""
    _validate_batch_limit(
        entity_name='metrics',
        limit=MAX_METRICS_PER_BATCH,
        length=len(metrics))
    _validate_batch_limit(
        entity_name='params',
        limit=MAX_PARAMS_TAGS_PER_BATCH,
        length=len(params))
    _validate_batch_limit(
        entity_name='tags', limit=MAX_PARAMS_TAGS_PER_BATCH, length=len(tags))
    total_length = len(metrics) + len(params) + len(tags)
    _validate_batch_limit(
        entity_name='metrics, params, and tags',
        limit=MAX_ENTITIES_PER_BATCH,
        length=total_length)


def _validate_batch_log_data(metrics, params, tags):
    for metric in metrics:
        _validate_metric(metric.key, metric.value, metric.timestamp,
                         metric.step)
        # TODO: move _validate_length_limit calls into _validate_metric etc.
        # This would be a breaking change as _validate_metric is also used in
        # the single-entry log_metric API. Thus we defer it for now to allow
        # for a release of the batched logging APIs without breaking
        # changes to other APIs. See related discussion in
        # https://github.com/mlflow/mlflow/issues/985
        _validate_length_limit('Metric name', MAX_ENTITY_KEY_LENGTH,
                               metric.key)
    for param in params:
        _validate_param(param.key, param.value)
    for tag in tags:
        _validate_tag(tag.key, tag.value)


def _validate_batch_log_api_req(json_req):
    if len(json_req) > MAX_BATCH_LOG_REQUEST_SIZE:
        error_msg = (
            'Batched logging API requests must be at most {limit} bytes, got '
            'a request of size {size}.').format(
                limit=MAX_BATCH_LOG_REQUEST_SIZE, size=len(json_req))
        raise MlflowException(error_msg, error_code=INVALID_PARAMETER_VALUE)


def _validate_experiment_name(experiment_name):
    """Check that `experiment_name` is a valid string and raise an exception if
    it isn't."""
    if experiment_name == '' or experiment_name is None:
        raise MlflowException(
            "Invalid experiment name: '%s'" % experiment_name,
            error_code=INVALID_PARAMETER_VALUE)

    if not is_string_type(experiment_name):
        raise MlflowException(
            'Invalid experiment name: %s. Expects a string.'
            % experiment_name, error_code=INVALID_PARAMETER_VALUE)


def _validate_experiment_artifact_location(artifact_location):
    if artifact_location is not None and artifact_location.startswith('runs:'):
        raise MlflowException(
            "Artifact location cannot be a runs:/ URI. Given: '%s'" %
            artifact_location,
            error_code=INVALID_PARAMETER_VALUE)


def _validate_db_type_string(db_type):
    # """validates db_type parsed from DB URI is supported."""
    if db_type not in DATABASE_ENGINES:  # noqa: F821
        error_msg = "Invalid database engine: '%s'. '%s'" % (
            db_type, _UNSUPPORTED_DB_TYPE_MSG)
        raise MlflowException(error_msg, INVALID_PARAMETER_VALUE)



from segmind.data.converters.constants import DEFAULT_ENCODING  # noqa: F401
from segmind.data.converters.constants import (FORMAT_PASCALVOC, FORMAT_YOLO,
                                               SETTING_ADVANCE_MODE,
                                               SETTING_AUTO_SAVE,
                                               SETTING_DRAW_SQUARE,
                                               SETTING_FILENAME,
                                               SETTING_FILL_COLOR,
                                               SETTING_LAST_OPEN_DIR,
                                               SETTING_LINE_COLOR,
                                               SETTING_PAINT_LABEL,
                                               SETTING_RECENT_FILES,
                                               SETTING_SAVE_DIR,
                                               SETTING_SINGLE_CLASS,
                                               SETTING_WIN_GEOMETRY,
                                               SETTING_WIN_POSE,
                                               SETTING_WIN_SIZE,
                                               SETTING_WIN_STATE)
from segmind.data.converters.pascal_voc_io import PascalVocReader  # noqa: F401
from segmind.data.converters.pascal_voc_io import PascalVocWriter  # noqa: F401
from segmind.data.converters.utils import \
    XMLReader  # noqa: F401; noqa: F401; noqa: F401; noqa: F401
from segmind.data.converters.utils import (check_argv, get, get_and_check,
                                           get_directory_xml_files,
                                           get_filename, process_file)
from segmind.data.converters.yolo_io import YoloReader  # noqa: F401
from segmind.data.converters.yolo_io import YOLOWriter



from segmind.entities._mlflow_object import _MLflowObject
from segmind.entities.metric import Metric
from segmind.entities.param import Param
from segmind.entities.run_tag import RunTag
from segmind.protos.service_lite_pb2 import Param as ProtoParam
from segmind.protos.service_lite_pb2 import RunData as ProtoRunData
from segmind.protos.service_lite_pb2 import RunTag as ProtoRunTag


class RunData(_MLflowObject):
    """Run data (metrics and parameters)."""

    def __init__(self, metrics=None, params=None, tags=None):
        """Construct a new :py:class:`cral.tracking.entities.RunData` instance.

        :param metrics: List of :py:class:`cral.tracking.entities.Metric`.
        :param params: List of :py:class:`cral.tracking.entities.Param`.
        :param tags: List of :py:class:`cral.tracking.entities.RunTag`.
        """
        # Maintain the original list of metrics so that we can easily convert
        # it back to protobuf
        self._metric_objs = metrics or []
        self._metrics = {
            metric.key: metric.value
            for metric in self._metric_objs
        }
        self._params = {param.key: param.value for param in (params or [])}
        self._tags = {tag.key: tag.value for tag in (tags or [])}

    @property
    def metrics(self):
        """Dictionary of string key -> metric value for the current run.

        For each metric key, the metric value with the latest timestamp is
        returned. In case there are multiple values with the same latest
        timestamp, the maximum of these values is returned.
        """
        return self._metrics

    @property
    def params(self):
        """Dictionary of param key (string) -> param value for the current
        run."""
        return self._params

    @property
    def tags(self):
        """Dictionary of tag key (string) -> tag value for the current run."""
        return self._tags

    def _add_metric(self, metric):
        self._metrics[metric.key] = metric.value
        self._metric_objs.append(metric)

    def _add_param(self, param):
        self._params[param.key] = param.value

    def _add_tag(self, tag):
        self._tags[tag.key] = tag.value

    def to_proto(self):
        run_data = ProtoRunData()
        run_data.metrics.extend([m.to_proto() for m in self._metric_objs])
        run_data.params.extend([
            ProtoParam(key=key, value=val) for key, val in self.params.items()
        ])
        run_data.tags.extend([
            ProtoRunTag(key=key, value=val) for key, val in self.tags.items()
        ])
        return run_data

    def to_dictionary(self):
        return {
            'metrics': self.metrics,
            'params': self.params,
            'tags': self.tags,
        }

    @classmethod
    def from_proto(cls, proto):
        run_data = cls()
        # iterate proto and add metrics, params, and tags
        for proto_metric in proto.metrics:
            run_data._add_metric(Metric.from_proto(proto_metric))
        for proto_param in proto.params:
            run_data._add_param(Param.from_proto(proto_param))
        for proto_tag in proto.tags:
            run_data._add_tag(RunTag.from_proto(proto_tag))
        return run_data



# import sys

from segmind.entities._mlflow_object import _MLflowObject
from segmind.protos.service_lite_pb2 import Param as ProtoParam


class Param(_MLflowObject):
    """Parameter object."""

    def __init__(self, key, value, tags=()):
        # if 'pyspark.ml' in sys.modules:
        #     import pyspark.ml.param
        #     if isinstance(key, pyspark.ml.param.Param):
        #         key = key.name
        #         value = str(value)
        self._key = key
        self._value = value
        self._tags = tags

    @property
    def key(self):
        """String key corresponding to the parameter name."""
        return self._key

    @property
    def value(self):
        """String value of the parameter."""
        return self._value

    @property
    def tags(self):
        """String value of the parameter."""
        return self._tags

    def to_proto(self):
        param = ProtoParam()
        param.key = self.key
        param.value = self.value
        param.tags.extend(self.tags)
        return param

    @classmethod
    def from_proto(cls, proto):
        return cls(proto.key, proto.value, proto.tags)



from abc import ABCMeta, abstractmethod


class RunContextProvider(object):
    """Abstract base class for context provider objects specifying custom tags
    at run-creation time (e.g. tags specifying the git repo with which the run
    is associated).

    When a run is created via the fluent ``segmind_track.start_run`` method,
    MLflow iterates through all registered RunContextProviders. For each
    context provider where ``in_context`` returns ``True``, MLflow calls the
    ``tags`` method on the context provider to compute context tags for the
    run. All context tags are then merged together and set on the newly-created
    run.
    """

    __metaclass__ = ABCMeta

    @abstractmethod
    def in_context(self):
        """Determine if MLflow is running in this context.

        :return: bool indicating if in this context
        """

    @abstractmethod
    def tags(self):
        """Generate context-specific tags.

        :return: dict of tags
        """



import json

from segmind.protos.errorcodes_pb2 import (BAD_REQUEST, ENDPOINT_NOT_FOUND,
                                           INTERNAL_ERROR,
                                           INVALID_PARAMETER_VALUE,
                                           INVALID_STATE, PERMISSION_DENIED,
                                           REQUEST_LIMIT_EXCEEDED,
                                           RESOURCE_ALREADY_EXISTS,
                                           RESOURCE_DOES_NOT_EXIST,
                                           TEMPORARILY_UNAVAILABLE, ErrorCode)

ERROR_CODE_TO_HTTP_STATUS = {
    ErrorCode.Name(INTERNAL_ERROR): 500,
    ErrorCode.Name(INVALID_STATE): 500,
    ErrorCode.Name(TEMPORARILY_UNAVAILABLE): 503,
    ErrorCode.Name(REQUEST_LIMIT_EXCEEDED): 429,
    ErrorCode.Name(ENDPOINT_NOT_FOUND): 404,
    ErrorCode.Name(RESOURCE_DOES_NOT_EXIST): 404,
    ErrorCode.Name(PERMISSION_DENIED): 403,
    ErrorCode.Name(BAD_REQUEST): 400,
    ErrorCode.Name(RESOURCE_ALREADY_EXISTS): 400,
    ErrorCode.Name(INVALID_PARAMETER_VALUE): 400
}


class MlflowException(Exception):
    """Generic exception thrown to surface failure information about external-
    facing operations.

    The error message associated with this exception may be exposed to clients
    in HTTP responses for debugging purposes. If the error text is sensitive,
    raise a generic `Exception` object instead.
    """

    def __init__(self, message, error_code=INTERNAL_ERROR, **kwargs):
        """
        Args:
            message: The message describing the error that occured. This will
            be included in the exception's serialized JSON representation.
            error_code: An appropriate error code for the error that occured;
            it will be included in the exception's serialized JSON
            representation. This should be one of the codes listed in the
            `cral.tracking.protos.databricks_pb2` proto.
            kwargs: Additional key-value pairs to include in the serialized
            JSON representation of the MlflowException.
        """
        try:
            self.error_code = ErrorCode.Name(error_code)
        except (ValueError, TypeError):
            self.error_code = ErrorCode.Name(INTERNAL_ERROR)
        self.message = message
        self.json_kwargs = kwargs
        super(MlflowException, self).__init__(message)

    def serialize_as_json(self):
        exception_dict = {
            'error_code': self.error_code,
            'message': self.message
        }
        exception_dict.update(self.json_kwargs)
        return json.dumps(exception_dict)

    def get_http_status_code(self):
        return ERROR_CODE_TO_HTTP_STATUS.get(self.error_code, 500)


class RestException(MlflowException):
    """Exception thrown on non 200-level responses from the REST API."""

    def __init__(self, json):
        error_code = json.get('error_code', ErrorCode.Name(INTERNAL_ERROR))
        message = '%s: %s' % (error_code, json['message'] if 'message' in json
                              else 'Response: ' + str(json))
        super(RestException, self).__init__(
            message, error_code=ErrorCode.Value(error_code))
        self.json = json


class ExecutionException(MlflowException):
    """Exception thrown when executing a project fails."""


class MlflowliteException(Exception):
    """docstring for MlflowliteException."""

    def __init__(self, message, error_code=INTERNAL_ERROR):
        """
        Args:
        message: The message describing the error that occured. This will be
        included in the exception's serialized JSON representation.
        error_code: An appropriate error code for the error that occured; it
        will be included in the exception's serialized JSON representation.
        This should be one of the codes listed in the
        `segmind.tracking.protos.databricks_pb2` proto.
        kwargs: Additional key-value pairs to include in the serialized JSON
        representation of the MlflowException.
        """
        try:
            self.error_code = ErrorCode.Name(error_code)
        except (ValueError, TypeError):
            self.error_code = ErrorCode.Name(INTERNAL_ERROR)
        message = '{}: {}'.format(self.error_code, message)
        # self.json_kwargs = kwargs
        super(MlflowliteException, self).__init__(message)


class MissingConfigException(MlflowException):
    """Exception thrown when expected configuration file/directory not
    found."""



#!/usr/bin/python

# import cv2
import glob
import json
import numpy as np
import os
import os.path
import shutil
import tempfile
import xml.etree.ElementTree as ET
from pascal_voc_writer import Writer
from pycocotools.coco import COCO

from segmind.data.converters.pascal_voc_io import (PascalVocReader,
                                                   PascalVocWriter)
from segmind.data.converters.utils import get, get_and_check, process_file
from segmind.data.converters.yolo_io import YoloReader, YOLOWriter

# TODO: @shreeram
# use glob.glob in place of os.listdir
# use os.path.join() in place of + "/" +

START_BOUNDING_BOX_ID = 1
PRE_DEFINED_CATEGORIES = {}


def coco_to_voc_bbox(bbox):

    assert isinstance(bbox, np.ndarray) and bbox.ndim == 2 and bbox.shape[
        1] == 4, f'bbox should be numpy of dimension (Nx4), got {bbox.shape}'

    bbox[:, 2] = bbox[:, 2] + bbox[:, 0]
    bbox[:, 3] = bbox[:, 3] + bbox[:, 1]

    return bbox


# def yolo_to_voc_bbox(image, bbox):
#
#     image_array = cv2.imread(image)
#     assert isinstance(bbox, np.ndarray) and bbox.ndim == 2 and bbox.shape[
#         1] == 4, 'bbox should be numpy of dimension (Nx4)'
#     height, width, c = image_array.shape
#
#     bbox_widths = bbox[:, 2] * width
#     bbox_heights = bbox[:, 3] * height
#
#     bbox[:, 0] = bbox[:, 0] * width - (bbox_widths / 2)
#     bbox[:, 1] = bbox[:, 1] * height - (bbox_heights / 2)
#     bbox[:, 2] = bbox[:, 0] + bbox_widths
#     bbox[:, 3] = bbox[:, 1] + bbox_heights
#
#     return bbox


# def yolo_to_voc(ann_dir,
#                 imgfolderpath,
#                 classfilepath,
#                 output_dir=tempfile.gettempdir()):
#     # TODO: @shreeram, make all argument names small
#     """arguments :
#
#     ann_dir : path to the annotations folder
#     imgfolderpath : path to the iamges folder
#     classfilepath : path to the class file (.txt file)
#     output_dir : path where the converted files are to be stored
#     """
#
#     directory = 'ConvertedPascalVOCFiles'
#     dirpath = os.path.join(output_dir, directory)
#     if os.path.exists(dirpath) and os.path.isdir(dirpath):
#         shutil.rmtree(dirpath)
#     os.mkdir(dirpath)
#
#     for file in glob.glob(os.path.join(ann_dir, '*')):
#         if file.endswith('.txt') and file != 'classes.txt':
#
#             split_name = os.path.basename(file)
#             print('Converting', split_name)
#
#             annotation_txt = os.path.splitext(split_name)[0]
#             imgPath = os.path.join(imgfolderpath, annotation_txt + '.jpg')
#
#             image = cv2.imread(imgPath)
#             imageShape = [image.shape[0], image.shape[1], image.shape[2]]
#             imgFolderName = os.path.basename(imgfolderpath)
#             imgFileName = os.path.basename(imgPath)
#
#             writer = PascalVocWriter(
#                 imgFolderName, imgFileName, imageShape, localImgPath=imgPath)
#
#             # Read YOLO files
#             txt_path = os.path.join(ann_dir, split_name)
#             Yolo_reader = YoloReader(txt_path, image, classfilepath)
#             shapes = Yolo_reader.getShapes()
#             num_of_box = len(shapes)
#
#             for i in range(num_of_box):
#                 label = shapes[i][0]
#                 xmin = shapes[i][1][0][0]
#                 ymin = shapes[i][1][0][1]
#                 x_max = shapes[i][1][2][0]
#                 y_max = shapes[i][1][2][1]
#
#                 writer.addBndBox(xmin, ymin, x_max, y_max, label, 0)
#
#         # Write the converted PascalVOC xml files into a new Directory
#             writer.save(
#                 targetFile=os.path.join(dirpath, annotation_txt + '.xml'))
#     return dirpath


# def voc_to_yolo(ann_dir,
#                 imgfolderpath,
#                 debug=False,
#                 output_dir=tempfile.gettempdir()):
#     # TODO: @shreeram, make all argument names small
#     """arguments :
#
#     ann_dir : path to the annotations folder
#     imgfolderpath : path to the iamges folder
#     output_dir : path where the converted files are to be stored
#     """
#
#     directory = 'ConverterdYolotxtFiles'
#     dirpath = os.path.join(output_dir, directory)
#     if os.path.exists(dirpath) and os.path.isdir(dirpath):
#         shutil.rmtree(dirpath)
#     os.mkdir(dirpath)
#     classes = []
#
#     # Search all pascal annotation (xml files) in this folder
#     for file in glob.glob(os.path.join(
#             ann_dir, '*')):  # use glob.glob in place of os.listdir
#         if file.endswith('.xml'):
#
#             split_name = os.path.basename(file)
#             if debug:
#                 print(f'Converting :: {split_name}')
#             annotation_xml = os.path.splitext(split_name)[0]
#
#             imagePath = os.path.join(imgfolderpath, annotation_xml + '.jpg')
#
#             image = cv2.imread(imagePath)
#             imageShape = [image.shape[0], image.shape[1], image.shape[2]]
#             imgFolderName = os.path.basename(imgfolderpath)
#             imgFileName = os.path.basename(imagePath)
#
#             writer = YOLOWriter(
#                 imgFolderName, imgFileName, imageShape, localImgPath=imagePath)
#
#             parser = ET.XMLParser(encoding='utf-8')
#             tree = ET.parse(
#                 os.path.join(ann_dir, str(split_name)), parser=parser)
#             root = tree.getroot()
#             for obj in root.findall('object'):
#                 if str(obj.find('name').text) not in classes:
#                     classes.append(str(obj.find('name').text))
#
#             # Read VOC file
#             filePath = os.path.join(ann_dir, split_name)
#             Voc_Reader = PascalVocReader(filePath, image)
#             shapes = Voc_Reader.getShapes()
#             num_boxes = len(shapes)
#
#             for i in range(num_boxes):
#                 label = classes.index(shapes[i][0])
#                 xmin = shapes[i][1][0][0]
#                 ymin = shapes[i][1][0][1]
#                 x_max = shapes[i][1][2][0]
#                 y_max = shapes[i][1][2][1]
#
#                 writer.addBndBox(xmin, ymin, x_max, y_max, label, 0)
#
#             writer.save(
#                 targetFile=os.path.join(dirpath, annotation_xml + '.txt'))
#
#             classFile = open(os.path.join(output_dir, 'classes.txt'), 'w')
#             for element in classes:
#                 classFile.write(element + '\n')
#     return dirpath


def coco_to_voc(ann_file, output_dir=tempfile.gettempdir()):
    """arguments :

    ann_file : path to the annotations file (.json file)
    output_dir : path where the converted files are to be stored
    """

    coco = COCO(ann_file)
    cats = coco.loadCats(coco.getCatIds())
    cat_idx = {}
    for c in cats:
        cat_idx[c['id']] = c['name']
    for img in coco.imgs:
        catIds = coco.getCatIds()
        annIds = coco.getAnnIds(imgIds=[img], catIds=catIds)
        if len(annIds) > 0:
            img_fname = coco.imgs[img]['file_name']
            image_fname_ls = img_fname.split('.')
            image_fname_ls[-1] = 'xml'
            label_fname = '.'.join(image_fname_ls)
            writer = Writer(img_fname, coco.imgs[img]['width'],
                            coco.imgs[img]['height'])
            anns = coco.loadAnns(annIds)
            for a in anns:
                bbox = a['bbox']
                bbox = [bbox[0], bbox[1], bbox[2] + bbox[0], bbox[3] + bbox[1]]
                bbox = [str(b) for b in bbox]
                catname = cat_idx[a['category_id']]
                writer.addObject(catname, bbox[0], bbox[1], bbox[2], bbox[3])
                writer.save(output_dir + '/' + label_fname)

    return output_dir


def voc_to_coco(ann_dir, debug=False, output_dir=tempfile.gettempdir()):
    """
    ann_dir : path to the annotations folder
    output_dir : path where the converted files are to be stored
    """

    categories = PRE_DEFINED_CATEGORIES
    bnd_id = START_BOUNDING_BOX_ID
    json_dict = {
        'images': [],
        'type': 'instances',
        'annotations': [],
        'categories': []
    }
    image_id = 1
    for file in glob.glob(os.path.join(
            ann_dir, '*')):  # use glob.glob in place of os.listdir
        if file.endswith('xml'):

            split_name = os.path.basename(file)
            if debug:
                print(f'Converting :: {split_name}')
            xml_f = os.path.join(ann_dir, split_name)
            tree = ET.parse(xml_f)
            root = tree.getroot()
            path = root.findtext('path')
            if path is None:
                filename = root.findtext('filename')
            else:
                filename = os.path.basename(path)

            size = get_and_check(root, 'size', 1)
            width = int(get_and_check(size, 'width', 1).text)
            height = int(get_and_check(size, 'height', 1).text)
            image = {
                'file_name': filename,
                'height': height,
                'width': width,
                'id': image_id
            }
            json_dict['images'].append(image)

            for obj in get(root, 'object'):

                category = get_and_check(obj, 'name', 1).text
                if category not in categories:
                    new_id = len(categories)
                    categories[category] = new_id
                category_id = categories[category]
                bndbox = get_and_check(obj, 'bndbox', 1)
                xmin = float(get_and_check(bndbox, 'xmin', 1).text)
                ymin = float(get_and_check(bndbox, 'ymin', 1).text)
                xmax = float(get_and_check(bndbox, 'xmax', 1).text)
                ymax = float(get_and_check(bndbox, 'ymax', 1).text)
                assert (xmax > xmin)
                assert (ymax > ymin)
                o_width = abs(xmax - xmin)
                o_height = abs(ymax - ymin)
                ann = {
                    'area': o_width * o_height,
                    'iscrowd': 0,
                    'image_id': image_id,
                    'bbox': [xmin, ymin, o_width, o_height],
                    'category_id': category_id,
                    'id': bnd_id,
                    'ignore': 0,
                    'segmentation': []
                }
                json_dict['annotations'].append(ann)
                bnd_id = bnd_id + 1

            for cate, cid in categories.items():
                cat = {'supercategory': 'none', 'id': cid, 'name': cate}
                json_dict['categories'].append(cat)
            image_id += 1

            json_file = 'converted_file.json'
            path = os.path.join(output_dir, json_file)
            json_fp = open(path, 'w')
            json_str = json.dumps(json_dict)
            json_fp.write(json_str)
            json_fp.close()

    return (os.path.join(output_dir, json_file))


def voc_to_kitti(ann_dir, output_dir=tempfile.gettempdir()):
    """
    ann_dir : path to the annotations folder
    output_dir : path where the converted files are to be stored
    """

    file_count = 0

    for file in glob.glob(os.path.join(ann_dir, '*')):

        split_name = os.path.basename(file)
        path = os.path.join(ann_dir, split_name)
        if process_file(path, output_dir):
            file_count += 1

    print('Conversion completed. {0} Files are processed'.format(file_count))


def kitti_to_voc(ann_dir, imgfolderpath, output_dir):
    """
    ann_dir : path to the annotations folder
    imgfolderpath : path to the iamges folder
    output_dir : path where the converted files are to be stored
    """

    filter_item = ['DontCare']
    for filename in glob.glob(os.path.join(ann_dir, '*')):

        split_name = os.path.basename(filename)
        annotation_kitti = os.path.splitext(split_name)[0]

        imagePath = os.path.join(imgfolderpath, annotation_kitti + '.jpg')

        img = cv2.imread(imagePath)
        imageShape = [img.shape[0], img.shape[1], img.shape[2]]
        img_name = os.path.basename(imagePath)

        voc_writer = PascalVocWriter(output_dir, img_name, imageShape)
        count = 0
        kitti_ant = open(ann_dir + '/' + split_name, 'r')
        for line in kitti_ant:
            item_list = line.split(' ')
            if item_list[0] not in filter_item:
                count += 1
                voc_writer.addBndBox(
                    (float(item_list[4])), (float(item_list[5])),
                    (float(item_list[6])), (float(item_list[7])), item_list[0],
                    0)

            else:

                img[(float(item_list[5])):(float(item_list[7])),
                    (float(item_list[4])):(float(item_list[6])), :] = 0

        if count > 0:
            voc_writer.save(
                targetFile=os.path.join(output_dir, annotation_kitti + '.xml'))



import GPUtil


def gpu_metrics():
    gpu_data = {}
    gpus = GPUtil.getGPUs()

    for i, gpu in enumerate(gpus):
        gpu_data[f'GPU_load_{i}'] = gpu.load
        gpu_data[f'GPU_Memory_util_{i}'] = gpu.memoryUtil

    return gpu_data



import os

TRACKING_URI = os.environ.get('SEGMIND_TRACKING_URL',
                              'https://cloud.segmind.com/track')
SEGMIND_API_URL = os.environ.get('SEGMIND_API_URL',
                                 'https://cloud.segmind.com/api')
SEGMIND_SPOT_URL = os.environ.get('SEGMIND_SPOT_URL',
                                  'https://api.spotprod.segmind.com')



from segmind.entities._mlflow_object import _MLflowObject
from segmind.entities.run_data import RunData
from segmind.entities.run_info import RunInfo
from segmind.exceptions import MlflowException
from segmind.protos.service_lite_pb2 import Run as ProtoRun


class Run(_MLflowObject):
    """Run object."""

    def __init__(self, run_info, run_data):
        if run_info is None:
            raise MlflowException('run_info cannot be None')
        self._info = run_info
        self._data = run_data

    @property
    def info(self):
        """The run metadata, such as the run id, start time, and status.

        :rtype: :py:class:`cral.tracking.entities.RunInfo`
        """
        return self._info

    @property
    def data(self):
        """The run data, including metrics, parameters, and tags.

        :rtype: :py:class:`cral.tracking.entities.RunData`
        """
        return self._data

    def to_proto(self):
        run = ProtoRun()
        run.info.MergeFrom(self.info.to_proto())
        if self.data:
            run.data.MergeFrom(self.data.to_proto())
        return run

    @classmethod
    def from_proto(cls, proto):
        return cls(
            RunInfo.from_proto(proto.info), RunData.from_proto(proto.data))

    def to_dictionary(self):
        run_dict = {
            'info': dict(self.info),
        }
        if self.data:
            run_dict['data'] = self.data.to_dictionary()
        return run_dict



import click
import logging
import sys

from segmind.store.artifact.artifact_repository_registry import \
    get_artifact_repository
from segmind.tracking import _get_store
from segmind.tracking.artifact_utils import _download_artifact_from_uri
from segmind.utils.proto_json_utils import message_to_json

_logger = logging.getLogger(__name__)


@click.group('artifacts')
def commands():
    """Upload, list, and download artifacts from an MLflow artifact repository.

    To manage artifacts for a run associated with a tracking server, set the
    MLFLOW_TRACKING_URI environment variable to the URL of the desired server.
    """


@commands.command('log-artifact')
@click.option(
    '--local-file', '-l', required=True, help='Local path to artifact to log')
@click.option(
    '--run-id',
    '-r',
    required=True,
    help='Run ID into which we should log the artifact.')
@click.option(
    '--artifact-path',
    '-a',
    help='If specified, we will log the artifact into this subdirectory of '
    + "the run's artifact directory.")
def log_artifact(local_file, run_id, artifact_path):
    """Log a local file as an artifact of a run, optionally within a run-
    specific artifact path.

    Run artifacts can be organized into directories, so you can place the
    artifact in a directory this way.
    """
    store = _get_store()
    artifact_uri = store.get_run(run_id).info.artifact_uri
    artifact_repo = get_artifact_repository(artifact_uri)
    artifact_repo.log_artifact(local_file, artifact_path)
    _logger.info('Logged artifact from local file %s to artifact_path=%s',
                 local_file, artifact_path)


@commands.command('log-artifacts')
@click.option(
    '--local-dir',
    '-l',
    required=True,
    help='Directory of local artifacts to log')
@click.option(
    '--run-id',
    '-r',
    required=True,
    help='Run ID into which we should log the artifact.')
@click.option(
    '--artifact-path',
    '-a',
    help='If specified, we will log the artifact into this subdirectory of '
    + "the run's artifact directory.")
def log_artifacts(local_dir, run_id, artifact_path):
    """Log the files within a local directory as an artifact of a run,
    optionally within a run-specific artifact path.

    Run artifacts can be organized into directories, so you can place the
    artifact in a directory this way.
    """
    store = _get_store()
    artifact_uri = store.get_run(run_id).info.artifact_uri
    artifact_repo = get_artifact_repository(artifact_uri)
    artifact_repo.log_artifacts(local_dir, artifact_path)
    _logger.info('Logged artifact from local dir %s to artifact_path=%s',
                 local_dir, artifact_path)


@commands.command('list')
@click.option('--run-id', '-r', required=True, help='Run ID to be listed')
@click.option(
    '--artifact-path',
    '-a',
    help="If specified, a path relative to the run's root directory to list.")
def list_artifacts(run_id, artifact_path):
    """Return all the artifacts directly under run's root artifact directory,
    or a sub-directory.

    The output is a JSON-formatted list.
    """
    artifact_path = artifact_path if artifact_path is not None else ''
    store = _get_store()
    artifact_uri = store.get_run(run_id).info.artifact_uri
    artifact_repo = get_artifact_repository(artifact_uri)
    file_infos = artifact_repo.list_artifacts(artifact_path)
    print(_file_infos_to_json(file_infos))


def _file_infos_to_json(file_infos):
    json_list = [
        message_to_json(file_info.to_proto()) for file_info in file_infos
    ]
    return '[' + ', '.join(json_list) + ']'


@commands.command('download')
@click.option('--run-id', '-r', help='Run ID from which to download')
@click.option(
    '--artifact-path',
    '-a',
    help="For use with Run ID: if specified, a path relative to the run's "
    'root directory to download')
@click.option(
    '--artifact-uri',
    '-u',
    help='URI pointing to the artifact file or artifacts directory; use as an '
    'alternative to specifying --run_id and --artifact-path')
def download_artifacts(run_id, artifact_path, artifact_uri):
    """Download an artifact file or directory to a local directory. The output
    is the name of the file or directory on the local disk.

    Either ``--run-id`` or ``--artifact-uri`` must be provided.
    """
    if run_id is None and artifact_uri is None:
        _logger.error(
            'Either ``--run-id`` or ``--artifact-uri`` must be provided.')
        sys.exit(1)

    if artifact_uri is not None:
        print(_download_artifact_from_uri(artifact_uri))
        return

    artifact_path = artifact_path if artifact_path is not None else ''
    store = _get_store()
    artifact_uri = store.get_run(run_id).info.artifact_uri
    artifact_repo = get_artifact_repository(artifact_uri)
    artifact_location = artifact_repo.download_artifacts(artifact_path)
    print(artifact_location)


if __name__ == '__main__':
    commands()



"""Internal package providing a Python CRUD interface to MLflow experiments,
runs, registered models, and model versions.

This is a lower level API than the :py:mod:`segmind_track.tracking.fluent`
module, and is exposed in the :py:mod:`segmind_track.tracking` module.
"""

import os
import time

from segmind.entities import ViewType
from segmind.lite_extensions.client_utils import get_host_uri
from segmind.protos.service_lite_pb2 import (Artifact, ArtifactTag, LogMetric,
                                             LogParam)
from segmind.store.tracking import SEARCH_MAX_RESULTS_DEFAULT
from segmind.store.tracking.rest_store import RestStore

tracking_uri = get_host_uri()
# tracking_uri = 'http://0.0.0.0:5000'


class MlflowClient(object):
    """Client of an MLflow Tracking Server that creates and manages experiments
    and runs, and of an MLflow Registry Server that creates and manages
    registered models and model versions.

    It's a thin wrapper around TrackingServiceClient and RegistryClient so
    there is a unified API but we can keep the implementation of the tracking
    and registry clients independent from each other.
    """

    def __init__(self, tracking_uri=None, registry_uri=None):
        """
        :Args
        tracking_uri: Address of local or remote tracking server. If not
                    provided, defaults to the service set by
                    ``segmind_track.tracking.set_tracking_uri``. See
        registry_uri: Address of local or remote model registry server. If not
                    provided, defaults to the service set by
                    ``segmind_track.tracking.set_tracking_uri``.
        """
        final_tracking_uri = tracking_uri
        self._registry_uri = final_tracking_uri
        # self._tracking_client = TrackingServiceClient(final_tracking_uri)
        self._tracking_client = RestStore()

    def get_run(self, run_id):
        """
        Fetch the run from backend store. The resulting
        :py:class:`Run <segmind_track.entities.Run>` contains a collection of
        run metadata -- :py:class:`RunInfo <segmind_track.entities.RunInfo>`,
        as well as a collection of run parameters, tags, and metrics --
        :py:class:`RunData <segmind_track.entities.RunData>`. In the case
        where multiple metrics with the same key are logged for the run, the
        :py:class:`RunData <segmind_track.entities.RunData>` contains the most
        recently logged value at the largest step for each metric.

        :param run_id: Unique identifier for the run.

        :return: A single :py:class:`segmind_track.entities.Run` object, if
                the run exists. Otherwise, raises an exception.
        """
        return self._tracking_client.get_run(run_id)

    def get_metric_history(self, run_id, key):
        """Return a list of metric objects corresponding to all values logged
        for a given metric.

        :param run_id: Unique identifier for run
        :param key: Metric name within the run

        :return: A list of :py:class:`segmind_track.entities.Metric` entities
        if logged, else empty list
        """
        return self._tracking_client.get_metric_history(run_id, key)

    def create_run(self, experiment_id, start_time=None, tags=None):
        """Create a :py:class:`segmind_track.entities.Run` object that can be
        associated with metrics, parameters, artifacts, etc. Unlike
        :py:func:`segmind_track.projects.run`, creates objects but does not run
        code. Unlike :py:func:`segmind_track.start_run`, does not change the
        "active run" used by :py:func:`segmind_track.log_param`.

        :param experiment_id: The ID of then experiment to create a run in.
        :param start_time: If not provided, use the current timestamp.
        :param tags: A dictionary of key-value pairs that are converted into
                     :py:class:`segmind_track.entities.RunTag` objects.
        :return: :py:class:`segmind_track.entities.Run` that was created.
        """
        return self._tracking_client.create_run(
            experiment_id=experiment_id, start_time=start_time, tags=tags)

    def list_run_infos(self,
                       experiment_id,
                       run_view_type=ViewType.ACTIVE_ONLY):
        """:return: List of :py:class:`segmind_track.entities.RunInfo`"""
        return self._tracking_client.list_run_infos(experiment_id,
                                                    run_view_type)

    def list_experiments(self, view_type=None):
        """
        :return: List of :py:class:`segmind_track.entities.Experiment`
        """
        return self._tracking_client.list_experiments(view_type)

    def get_experiment(self, experiment_id):
        """Retrieve an experiment by experiment_id from the backend store.

        :Args
        experiment_id: The experiment ID returned from``create_experiment``.
        :return: :py:class:`segmind_track.entities.Experiment`
        """
        return self._tracking_client.get_experiment(experiment_id)

    # TO DO: @pk00095
    # Remove this
    def get_experiment_by_name(self, name):
        """Retrieve an experiment by experiment name from the backend store.

        :param name: The experiment name.
        :return: :py:class:`segmind_track.entities.Experiment`
        """
        return self._tracking_client.get_experiment_by_name(name)

    def create_experiment(self, name, artifact_location=None):
        """Create an experiment.

        :Args
        name: The experiment name. Must be unique.
        artifact_location: The location to store run artifacts.
                          If not provided, the server picks an appropriate
                          default.
        :return: Integer ID of the created experiment.
        """
        return self._tracking_client.create_experiment(name, artifact_location)

    def delete_experiment(self, experiment_id):

        # TO DO: @pk00095 # Remove this
        """Delete an experiment from the backend store.

        :Args
        experiment_id: The experiment ID returned from
                        ``create_experiment``.
        """
        self._tracking_client.delete_experiment(experiment_id)

    def restore_experiment(self, experiment_id):
        # TO DO: @pk00095 # Remove this
        """Restore a deleted experiment unless permanently deleted.

        :Args
        experiment_id: The experiment ID returned from ``create_experiment``.
        """
        self._tracking_client.restore_experiment(experiment_id)

    def rename_experiment(self, experiment_id, new_name):
        # TO DO: @pk00095 # Remove this
        """Update an experiment's name. The new name must be unique.

        :Args
        experiment_id: The experiment ID returned from ``create_experiment``.
        """
        self._tracking_client.rename_experiment(experiment_id, new_name)

    def log_metric(
            self,
            run_id,
            key,
            value,
            timestamp=None,
            step=None,
            epoch=None,
            tags=()):
        """Log a metric against the run ID.

        :Args
        run_id: The run id to which the metric should be logged.
        key: Metric name.
        value: Metric value (float). Note that some special values such
              as +/- Infinity may be replaced by other values depending on the
              store. For example, the SQLAlchemy store replaces +/- Inf with
              max / min float values.
        timestamp: Time when this metric was calculated. Defaults to the
                    current system time.
        step: Training step (iteration) at which was the metric calculated.
            Defaults to 0.
        """
        metric = LogMetric(
            key=str(key),
            value=value,
            timestamp=timestamp,
            step=step,
            epoch=epoch,
            tags=tags)
        self._tracking_client.log_metric(run_id, metric)

    def log_param(self, run_id, key, value, tags=()):
        """Log a parameter against the run ID.

        Value is converted to a string.
        """
        param = LogParam(
            key=str(key),
            value=str(value),
            tags=tags)
        self._tracking_client.log_param(run_id, param)

    def log_artifact_lite(self,
                          run_id,
                          experiment_id,
                          key,
                          path,
                          artifact_type='data',
                          timestamp=None,
                          size=None,
                          prediction=None,
                          ground_truth=None,
                          step=None,
                          tags={}):
        """Log an artifact against the run ID.

        :Args
        run_id: The run ID against which the artifact should be logged.
        experiment_id: The experiment ID against the run ID.
        key: Artifact name.
        path: Artifact path. This value should be the absolute path of the
            local artifact.
        """
        timestamp = timestamp or int(time.time() * 1000)
        size = os.path.getsize(path)

        tags_list = []
        for key, value in tags.items():
            tag = ArtifactTag(key=str(key), value=str(value))
            tags_list.append(tag)
        artifact = Artifact(
            key=key,
            path=path,
            artifact_type=artifact_type,
            timestamp=timestamp,
            size=size,
            prediction=prediction,
            ground_truth=ground_truth,
            step=step,
            tags=tags_list)
        self._tracking_client.log_artifact(run_id, experiment_id, artifact)

    def set_project_tag(self, experiment_id, key, value):
        """Set a tag on the experiment with the specified ID. Value is
        converted to a string.

        :param experiment_id: String ID of the experiment.
        :param key: Name of the tag.
        :param value: Tag value (converted to a string).
        """

        self._tracking_client.set_project_tag(experiment_id, key, value)

    def set_tag(self, run_id, key, value):
        """Set a tag on the run with the specified ID. Value is converted to a
        string.

        :param run_id: String ID of the run.
        :param key: Name of the tag.
        :param value: Tag value (converted to a string)
        """
        self._tracking_client.set_tag(run_id, key, value)

    def delete_tag(self, run_id, key):
        """Delete a tag from a run. This is irreversible.

        :param run_id: String ID of the run
        :param key: Name of the tag
        """
        self._tracking_client.delete_tag(run_id, key)

    def log_batch(self, run_id, metrics=(), params=(), tags=()):
        """Log multiple metrics, params, and/or tags.

        :Args
        run_id: String ID of the run
        metrics: If provided, List of Metric(key, value, timestamp) instances.
        params: If provided, List of Param(key, value) instances.
        tags: If provided, List of RunTag(key, value) instances.

        Raises an MlflowException if any errors occur.
        :return: None
        """
        self._tracking_client.log_batch(run_id, metrics, params, tags)

    def log_artifact(self, run_id, local_path, artifact_path=None):
        """Write a local file or directory to the remote ``artifact_uri``.

        :Args
        local_path: Path to the file or directory to write.
        artifact_path: If provided, the directory in ``artifact_uri`` to
                    write to.
        """
        self._tracking_client.log_artifact(run_id, local_path, artifact_path)

    def log_artifacts(self, run_id, local_dir, artifact_path=None):
        """Write a directory of files to the remote ``artifact_uri``.

        :Args
        local_dir: Path to the directory of files to write.
        artifact_path: If provided, the directory in ``artifact_uri`` to write
                        to.
        """
        self._tracking_client.log_artifacts(run_id, local_dir, artifact_path)

    def _record_logged_model(self, run_id, mlflow_model):
        """Record logged model info with the tracking server.

        :Args
        run_id: run_id under which the model has been logged.
        mlflow_model: Model info to be recorded.
        """
        self._tracking_client._record_logged_model(run_id, mlflow_model)

    def list_artifacts(self, run_id, path=None):
        """List the artifacts for a run.

        :Args
        run_id: The run to list artifacts from.
        path: The run's relative artifact path to list from. By default it is
            set to None or the root artifact path.
        :return: List of :py:class:`segmind_track.entities.FileInfo`
        """
        return self._tracking_client.list_artifacts(run_id, path)

    def download_artifacts(self, run_id, path, dst_path=None):
        """Download an artifact file or directory from a run to a local
        directory if applicable, and return a local path for it.

        :Args
        run_id: The run to download artifacts from.
        path: Relative source path to the desired artifact.
        dst_path: Absolute path of the local filesystem destination directory
                to which to download the specified artifacts. This directory
                must already exist. If unspecified, the artifacts will either
                be downloaded to a new uniquely-named directory on the local
                filesystem or will be returned directly in the case of the
                LocalArtifactRepository.
        :return: Local path of desired artifact.
        """
        return self._tracking_client.download_artifacts(run_id, path, dst_path)

    def set_terminated(self, run_id, status=None, end_time=None):
        """Set a run's status to terminated.

        :Args
        status: A string value of :py:class:`segmind_track.entities.RunStatus`
               Defaults to "FINISHED".
        end_time: If not provided, defaults to the current time.
        """
        self._tracking_client.set_terminated(run_id, status, end_time)

    def delete_run(self, run_id):
        """Deletes a run with the given ID."""
        self._tracking_client.delete_run(run_id)

    def restore_run(self, run_id):
        """Restores a deleted run with the given ID."""
        self._tracking_client.restore_run(run_id)

    def search_runs(self,
                    experiment_ids,
                    filter_string='',
                    run_view_type=ViewType.ACTIVE_ONLY,
                    max_results=SEARCH_MAX_RESULTS_DEFAULT,
                    order_by=None,
                    page_token=None):
        """Search experiments that fit the search criteria.

        :Args
        experiment_ids: List of experiment IDs, or a single int or string id.
        filter_string: Filter query string, defaults to searching all runs.
        run_view_type: one of enum values ACTIVE_ONLY, DELETED_ONLY, or ALL
                    runs defined in :py:class:`segmind_track.entities.ViewType`. # noqa
        max_results: Maximum number of runs desired.
        order_by: List of columns to order by (e.g., "metrics.rmse"). The
                ``order_by`` column can contain an optional ``DESC`` or
                ``ASC`` value. The default is ``ASC``. The default ordering is
                to sort by ``start_time DESC``, then ``run_id``.
        page_token: Token specifying the next page of results. It should be
                obtained from a ``search_runs`` call.

        :return: A list of :py:class:`segmind_track.entities.Run` objects that
                satisfy the search expressions. If the underlying tracking
                store supports pagination, the token for the next page may be
                obtained via the ``token`` attribute of the returned object.
        """
        return self._tracking_client.search_runs(experiment_ids, filter_string,
                                                 run_view_type, max_results,
                                                 order_by, page_token)

    def create_registered_model(self, name):
        """Create a new registered model in backend store.

        :Args
        name: Name of the new model. This is expected to be unique in the
            backend store.
        :return: A single object of :py:class:`segmind_track.entities.model_registry.RegisteredModel` # noqa
                 created by backend.
        """
        return self._get_registry_client().create_registered_model(name)

    def update_registered_model(self, name, new_name=None, description=None):
        """Updates metadata for RegisteredModel entity. Either ``new_name`` or
        ``description`` should be non-None. Backend raises exception if a
        registered model with given name does not exist.

        :Args
        name: Name of the registered model to update.
        new_name: (Optional) New proposed name for the registered model.
        description: (Optional) New description.
        :return: A single updated :py:class:`segmind_track.entities.model_registry.RegisteredModel` object. # noqa
        """
        return self._get_registry_client().update_registered_model(
            name, new_name, description)

    def delete_registered_model(self, name):
        """Delete registered model. Backend raises exception if a registered
        model with given name does not exist.

        :param name: Name of the registered model to update.
        """
        self._get_registry_client().delete_registered_model(name)

    def list_registered_models(self):
        """List of all registered models.

        :return: List of :py:class:`segmind_track.entities.model_registry.RegisteredModel` objects. # noqa
        """
        return self._get_registry_client().list_registered_models()

    def get_registered_model_details(self, name):
        """

        :Args
        name: Name of the registered model to update.
        :return: A single :py:class:`segmind_track.entities.model_registry.RegisteredModelDetailed` object. # noqa
        """
        return self._get_registry_client().get_registered_model_details(name)

    def get_latest_versions(self, name, stages=None):
        """Latest version models for each requests stage. If no ``stages``
        provided, returns the latest version for each stage.

        :param name: Name of the registered model to update.
        :param stages: List of desired stages. If input list is None, return latest versions for
                       for ALL_STAGES.
        :return: List of `:py:class:`segmind_track.entities.model_registry.ModelVersionDetailed` objects. # noqa
        """
        return self._get_registry_client().get_latest_versions(name, stages)

    def create_model_version(self, name, source, run_id):
        """Create a new model version from given source or run ID.

        :param name: Name ID for containing registered model.
        :param source: Source path where the MLflow model is stored.
        :param run_id: Run ID from MLflow tracking server that generated the model
        :return: Single :py:class:`segmind_track.entities.model_registry.ModelVersion` object created by # noqa
                 backend.
        """
        return self._get_registry_client().create_model_version(
            name, source, run_id)

    def update_model_version(self,
                             name,
                             version,
                             stage=None,
                             description=None):
        """Update metadata associated with a model version in backend.

        :param name: Name of the containing registered model.
        :param version: Version number of the model version.
        :param stage: New desired stage for this model version.
        :param description: New description.
        """
        self._get_registry_client().update_model_version(
            name, version, stage, description)

    # @experimental
    def delete_model_version(self, name, version):
        """Delete model version in backend.

        :param name: Name of the containing registered model.
        :param version: Version number of the model version.
        """
        self._get_registry_client().delete_model_version(name, version)

    def get_model_version_details(self, name, version):
        """
        :param name: Name of the containing registered model.
        :param version: Version number of the model version.
        :return: A single :py:class:`segmind_track.entities.model_registry.ModelVersionDetailed` object. # noqa
        """
        return self._get_registry_client().get_model_version_details(
            name, version)

    def get_model_version_download_uri(self, name, version):
        """Get the download location in Model Registry for this model version.

        :param name: Name of the containing registered model.
        :param version: Version number of the model version.
        :return: A single URI location that allows reads for downloading.
        """
        return self._get_registry_client().get_model_version_download_uri(
            name, version)

    def search_model_versions(self, filter_string):
        """Search for model versions in backend that satisfy the filter
        criteria.

        :Args
        filter_string: A filter string expression. Currently supports a single
                    filter condition either name of model like
                    ``name = 'model_name'`` or ``run_id = '...'``.
        :return: PagedList of :py:class:`segmind_track.entities.model_registry.ModelVersion` objects. # noqa
        """
        return self._get_registry_client().search_model_versions(filter_string)

    def get_model_version_stages(self, name, version):
        """
        :return: A list of valid stages.
        """
        return self._get_registry_client().get_model_version_stages(
            name, version)



from segmind.sys_data import gpu_metrics, system_metrics
from segmind.tracking.fluent import log_metrics
from segmind.utils.logging_utils import try_mlflow_log


def XGBoost_callback(period=1):
    """Create a callback that print evaluation result.
    We print the evaluation results every **period** iterations
    and on the first and the last iterations.
    Parameters
    ----------
    period : int
        The period to log the evaluation results
    show_stdv : bool, optional
         Whether show stdv if provided
    Returns
    -------
    callback : function
        A callback that print evaluation every period iterations.
    """
    def callback(env):
        """internal function."""
        if env.rank != 0 or (not env.evaluation_result_list) or period is False or period == 0:  # noqa: E501
            return
        step = env.iteration

        results = {}
        # gpu_data = gpu_metrics()
        # results.update(gpu_data)

        # Removing system_metrics for now, as these are not frequently used
        # cpu_data = system_metrics()
        # results.update(cpu_data)
        if step % period == 0 or step + 1 == env.begin_iteration or step + 1 == env.end_iteration:  # noqa: E501
            for x in env.evaluation_result_list:
                results[x[0]] = x[1]
            try_mlflow_log(log_metrics, results, step=step)

    return callback



"""The ``segmind_track.tracking`` module provides a Python CRUD interface to
MLflow experiments and runs.

This is a lower level API that directly translates to MLflow `REST API
<../rest-api.html>`_ calls. For a higher level API for managing an "active
run", use the :py:mod:`mlflow` module.
"""

from segmind.tracking._tracking_service.utils import (_TRACKING_URI_ENV_VAR,
                                                      _get_store,
                                                      get_tracking_uri,
                                                      is_tracking_uri_set,
                                                      set_tracking_uri)
from segmind.tracking.client import MlflowClient
from segmind.tracking.fluent import _EXPERIMENT_ID_ENV_VAR, _RUN_ID_ENV_VAR
from segmind.tracking.fluent import set_project, create_experiment, log_param, log_params, log_metric, log_metrics, \
    log_artifact, log_image, log_table, set_runid, start_run, active_run, end_run, search_runs, get_run, delete_run
from segmind.utils.logging_utils import log_params_decorator

__all__ = [
    'MlflowClient',
    'get_tracking_uri',
    'set_tracking_uri',
    'is_tracking_uri_set',
    '_get_store',
    '_EXPERIMENT_ID_ENV_VAR',
    '_RUN_ID_ENV_VAR',
    '_TRACKING_URI_ENV_VAR',

    # tracking endpoints
    'set_project',
    'create_experiment',
    'start_run',
    'set_runid',
    'log_param',
    'log_params',
    'log_params_decorator',
    'log_metric',
    'log_metrics',
    'log_artifact',
    'log_image',
    'log_table',
    'active_run',
    'end_run',
    'search_runs',
    'get_run',
    'delete_run',
]



from abc import ABCMeta, abstractmethod

from segmind.entities import ViewType
from segmind.store.entities.paged_list import PagedList
from segmind.store.tracking import SEARCH_MAX_RESULTS_DEFAULT


class AbstractStore:
    """Abstract class for Backend Storage.

    This class defines the API interface for front ends to connect with various
    types of backends.
    """

    __metaclass__ = ABCMeta

    def __init__(self):
        """Empty constructor for now.

        This is deliberately not marked as abstract, else every derived class
        would be forced to create one.
        """

    @abstractmethod
    def list_experiments(self, view_type=ViewType.ACTIVE_ONLY):
        """
        :param view_type: Qualify requested type of experiments.

        :return: a list of Experiment objects stored in store for requested
        view.
        """

    @abstractmethod
    def create_experiment(self, name, artifact_location):
        """Create a new experiment. If an experiment with the given name
        already exists, throws exception.

        name: Desired name for an experiment
        artifact_location: Base location for artifacts in runs. May be None.

        :return: experiment_id (string) for the newly created experiment if
        successful, else None.
        """

    @abstractmethod
    def get_experiment(self, experiment_id):
        """Fetch the experiment by ID from the backend store.

        :param experiment_id: String id for the experiment

        :return: A single :py:class:`segmind_track.entities.Experiment` object
            if it exists, otherwise raises an exception.
        """

    def get_experiment_by_name(self, experiment_name):
        """Fetch the experiment by name from the backend store. This is a base
        implementation using ``list_experiments``, derived classes may have
        some specialized implementations.

        :param experiment_name: Name of experiment

        :return: A single :py:class:`segmind_track.entities.Experiment` object
                if it exists.
        """
        for experiment in self.list_experiments(ViewType.ALL):
            if experiment.name == experiment_name:
                return experiment
        return None

    @abstractmethod
    def delete_experiment(self, experiment_id):
        """Delete the experiment from the backend store. Deleted experiments
        can be restored until permanently deleted.

        :param experiment_id: String id for the experiment
        """

    @abstractmethod
    def restore_experiment(self, experiment_id):
        """Restore deleted experiment unless it is permanently deleted.

        :param experiment_id: String id for the experiment
        """

    @abstractmethod
    def rename_experiment(self, experiment_id, new_name):
        """Update an experiment's name. The new name must be unique.

        :param experiment_id: String id for the experiment
        """

    @abstractmethod
    def get_run(self, run_id):
        """Fetch the run from backend store. The resulting :py:class:`Run.

        <segmind_track.entities.Run>`

        contains a collection of run metadata -
        :py:class:`RunInfo <segmind_track.entities.RunInfo>`, as well as a
        collection of run parameters, tags, and metrics -
        :py:class`RunData <segmind_track.entities.RunData>`. In the case where
        multiple metrics with the same key are logged for the run, the
        :py:class:`RunData <segmind_track.entities.RunData>` contains the
        value at the latest timestamp for each metric. If there are multiple
        values with the latest timestamp for a given metric, the maximum of
        these values is returned.

        :param run_id: Unique identifier for the run.

        :return: A single :py:class:`segmind_track.entities.Run` object, if
                the run exists. Otherwise, raises an exception.
        """

    @abstractmethod
    def update_run_info(self, run_id, run_status, end_time):
        """Update the metadata of the specified run.

        :return: :py:class:`segmind_track.entities.RunInfo` describing the
                updated run.
        """

    @abstractmethod
    def create_run(self, experiment_id, user_id, start_time, tags):
        """Create a run under the specified experiment ID, setting the run's
        status to "RUNNING" and the start time to the current time.

        :param experiment_id: String id of the experiment for this run
        :param user_id: ID of the user launching this run

        :return: The created Run object
        """

    @abstractmethod
    def delete_run(self, run_id):
        """Delete a run.

        :param run_id
        """

    @abstractmethod
    def restore_run(self, run_id):
        """Restore a run.

        :param run_id
        """

    def log_metric(self, run_id, metric):
        """Log a metric for the specified run.

        run_id: String id for the run
        metric: :py:class:`segmind_track.entities.Metric` instance to log
        """
        self.log_batch(run_id, metrics=[metric], params=[], tags=[])

    def log_param(self, run_id, param):
        """Log a param for the specified run.

        :param run_id: String id for the run
        :param param: :py:class:`segmind_track.entities.Param` instance to log
        """
        self.log_batch(run_id, metrics=[], params=[param], tags=[])

    def log_artifact(self, run_id, artifact):
        """Log an artifact for the specified run.

        run_id: String id for the run
        artifact: :py:class:`segmind_track.entities.Artifact` instance to log
        """

    def set_project_tag(self, experiment_id, tag):
        """Set a tag for the specified experiment.

        experiment_id: String id for the experiment
        tag: :py:class:`segmind_track.entities.ExperimentTag` instance to set
        """

    def set_tag(self, run_id, tag):
        """Set a tag for the specified run.

        run_id: String id for the run
        tag: :py:class:`segmind_track.entities.RunTag` instance to set
        """
        self.log_batch(run_id, metrics=[], params=[], tags=[tag])

    @abstractmethod
    def get_metric_history(self, run_id, metric_key):
        """Return a list of metric objects corresponding to all values logged
        for a given metric.

        run_id: Unique identifier for run
        metric_key: Metric name within the run

        :return: A list of :py:class:`segmind_track.entities.Metric` entities
                if logged, else empty list
        """

    def search_runs(self,
                    experiment_ids,
                    filter_string,
                    run_view_type,
                    max_results=SEARCH_MAX_RESULTS_DEFAULT,
                    order_by=None,
                    page_token=None):
        """Return runs that match the given list of search expressions within
        the experiments.

        page_token:
        page_token:
        experiment_ids: List of experiment ids to scope the search
        filter_string: A search filter string.
        run_view_type: ACTIVE_ONLY, DELETED_ONLY, or ALL runs
        max_results: Maximum number of runs desired.
        order_by: List of order_by clauses.
        page_token: Token specifying the next page of results. It should be
                    obtained from a ``search_runs`` call.

        :return: A list of :py:class:`segmind_track.entities.Run` objects that
                satisfy the search expressions. The pagination token for the
                next page can be obtained via the ``token`` attribute of the
                object; however, some store implementations may not support
                pagination and thus the returned token would not be meaningful
                in such cases.
        """
        runs, token = self._search_runs(experiment_ids, filter_string,
                                        run_view_type, max_results, order_by,
                                        page_token)
        return PagedList(runs, token)

    @abstractmethod
    def _search_runs(self, experiment_ids, filter_string, run_view_type,
                     max_results, order_by, page_token):
        """Return runs that match the given list of search expressions within
        the experiments, as well as a pagination token (indicating where the
        next page should start). Subclasses of ``AbstractStore`` should
        implement this method to support pagination instead of ``search_runs``.

        See ``search_runs`` for parameter descriptions.

        :return: A tuple of ``runs`` and ``token`` where ``runs`` is a list of
            :py:class:`segmind_track.entities.Run` objects that satisfy the
            search expressions, and ``token`` is the pagination token for the
            next page of results.
        """

    def list_run_infos(self, experiment_id, run_view_type):
        """Return run information for runs which belong to the experiment_id.

        :param experiment_id: The experiment id which to search

        :return: A list of :py:class:`segmind_track.entities.RunInfo` objects
                that satisfy the search expressions
        """
        runs = self.search_runs([experiment_id], None, run_view_type)
        return [run.info for run in runs]

    @abstractmethod
    def log_batch(self, run_id, metrics, params, tags):
        """Log multiple metrics, params, and tags for the specified run.

        run_id: String id for the run
        metrics: List of :py:class:`segmind_track.entities.Metric` instances
                to log
        params: List of :py:class:`segmind_track.entities.Param` instances to
                log
        tags: List of :py:class:`segmind_track.entities.RunTag` instances to
                log

        :return: None.
        """

    @abstractmethod
    def record_logged_model(self, run_id, mlflow_model):
        """Record logged model information with tracking store. The list of
        logged model infos is maintained in a segmind_track.models tag in JSON
        format.

        Note: The actual models are logged as artifacts via artifact
            repository.

        :param run_id: String id for the run
        :param mlflow_model: Model object to be recorded.

        NB: This API is experimental and may change in the future. The default
        implementation is a no-op.

        :return: None.
        """



import os

_SEGMIND_FOLDER = os.path.join(os.path.expanduser('~'), '.segmind')
os.makedirs(_SEGMIND_FOLDER, exist_ok=True)

_PROJECTS_FOLDER = os.path.join(
    os.path.expanduser('~'), '.segmind', 'projects')
os.makedirs(_PROJECTS_FOLDER, exist_ok=True)


def red_print(text):
    print('\033[31m{}\033[0m'.format(text))


def cyan_print(text):
    print('\033[36m{}\033[0m'.format(text))


def green_print(text):
    print('\033[32m{}\033[0m'.format(text))


def yellow_print(text):
    print('\033[33m{}\033[0m'.format(text))


def blue_print(text):
    print('\033[34m{}\033[0m'.format(text))


def create_new_project(project_name):

    if os.path.isfile(project_name):
        raise ValueError('file already exists {}'.format(project_name))

    with open(project_name, 'w') as f:  # noqa: F841
        pass



import logging
import os

from segmind.tracking.context.abstract_context import RunContextProvider
from segmind.tracking.context.default_context import _get_main_file
from segmind.utils.mlflow_tags import MLFLOW_GIT_COMMIT

_logger = logging.getLogger(__name__)


def _get_git_commit(path):
    try:
        import git
    except ImportError as e:
        _logger.warning(
            'Failed to import Git (the Git executable is probably not on your'
            ' PATH), so Git SHA is not available. Error: %s', e)
        return None
    try:
        if os.path.isfile(path):
            path = os.path.dirname(path)
        repo = git.Repo(path, search_parent_directories=True)
        commit = repo.head.commit.hexsha
        return commit
    except (git.InvalidGitRepositoryError, git.GitCommandNotFound, ValueError,
            git.NoSuchPathError):
        return None


def _get_source_version():
    main_file = _get_main_file()
    if main_file is not None:
        return _get_git_commit(main_file)
    return None


class GitRunContext(RunContextProvider):

    def __init__(self):
        self._cache = {}

    @property
    def _source_version(self):
        if 'source_version' not in self._cache:
            self._cache['source_version'] = _get_source_version()
        return self._cache['source_version']

    def in_context(self):
        return self._source_version is not None

    def tags(self):
        return {MLFLOW_GIT_COMMIT: self._source_version}



# import numpy as np
# import pandas as pd
# import unittest
#
#
# class LogSpecialArtifacts(unittest.TestCase):
#     """docstring for log_images."""
#
#     def setUp(self):
#         pass
#         # from segmind import set_project
#         # set_project('f5082153-3ed2-40b6-a4c6-f4cd96b6cedc')
#
#     def test_LogImage(self):
#         from segmind import log_image
#
#         log_image(
#             key='panda.jpg',
#             image='test_data/Kung_fu_panda_poster.jpg',
#             step=0)
#
#     def test_LogObjectdetection(self):
#         from segmind import log_bbox_prediction
#
#         bboxes = [[
#             1358.99999999904, 573.00000000021, 1570.9999999996799,
#             620.00000000019
#         ],
#                   [
#                       535.9999999999679, 473.999999999472, 622.0000000000321,
#                       511.999999999488
#                   ],
#                   [
#                       1193.9999999993279, 675.0000000003779, 1270.999999999392,
#                       712.000000000422
#                   ],
#                   [
#                       1188.0000000006721, 774.000000000252, 1237.000000000608,
#                       852.0000000002279
#                   ],
#                   [
#                       1700.9999999993279, 581.0000000003639, 1783.999999999392,
#                       616.000000000356
#                   ],
#                   [
#                       1603.000000000032, 584.000000000496, 1681.999999999968,
#                       616.000000000464
#                   ],
#                   [
#                       1523.00000000064, 644.000000000004, 1595.00000000064,
#                       678.999999999996
#                   ]]
#
#         log_bbox_prediction(
#             key='DJI_0005-0151bbox_only.jpg',
#             image='test_data/object_detection/DJI_0005-0151.jpg',
#             bbox_pred=bboxes,
#             bbox_gt=None,
#             bbox_type='pascal_voc',
#             step=None)
#
#     # def test_LogSemanticMask(self):
#     #     from segmind import log_mask_prediction
#
#     #     log_mask_prediction(
#     #       key='doggy',
#     #       image='test_data/semantic_segmentation/dog2.jpg',
#     #       pred_mask='test_data/semantic_segmentation/dog2.png',
#     #       bbox_type='pascal_voc',
#     #       step=100)
#
#
# class LogTables(unittest.TestCase):
#     """docstring for log_tables."""
#
#     def setUp(self):
#         pass
#         # from segmind import set_project
#         # set_project('f5082153-3ed2-40b6-a4c6-f4cd96b6cedc')
#
#     def test_Pandas(self):
#
#         from segmind import log_table
#
#         data = {
#             'Int-Col': np.random.randint(low=2, size=10),
#             'Float-Col': np.random.rand(10)
#         }
#
#         dataframe = pd.DataFrame.from_dict(data)
#
#         log_table(key='test_table', table=dataframe, step=10)


if __name__ == '__main__':
    # unittest.main(verbosity=2)
    pass



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: scalapb/scalapb.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import descriptor_pb2 as google_dot_protobuf_dot_descriptor__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='scalapb/scalapb.proto',
  package='scalapb',
  syntax='proto2',
  serialized_options=_b('\n#org.cral.tracking.scalapb_interface'),
  serialized_pb=_b('\n\x15scalapb/scalapb.proto\x12\x07scalapb\x1a google/protobuf/descriptor.proto\"L\n\x0eScalaPbOptions\x12\x14\n\x0cpackage_name\x18\x01 \x01(\t\x12\x14\n\x0c\x66lat_package\x18\x02 \x01(\x08\x12\x0e\n\x06import\x18\x03 \x03(\t\"!\n\x0eMessageOptions\x12\x0f\n\x07\x65xtends\x18\x01 \x03(\t\"\x1c\n\x0c\x46ieldOptions\x12\x0c\n\x04type\x18\x01 \x01(\t:G\n\x07options\x12\x1c.google.protobuf.FileOptions\x18\xfc\x07 \x01(\x0b\x32\x17.scalapb.ScalaPbOptions:J\n\x07message\x12\x1f.google.protobuf.MessageOptions\x18\xfc\x07 \x01(\x0b\x32\x17.scalapb.MessageOptions:D\n\x05\x66ield\x12\x1d.google.protobuf.FieldOptions\x18\xfc\x07 \x01(\x0b\x32\x15.scalapb.FieldOptionsB%\n#org.cral.tracking.scalapb_interface')
  ,
  dependencies=[google_dot_protobuf_dot_descriptor__pb2.DESCRIPTOR,])


OPTIONS_FIELD_NUMBER = 1020
options = _descriptor.FieldDescriptor(
  name='options', full_name='scalapb.options', index=0,
  number=1020, type=11, cpp_type=10, label=1,
  has_default_value=False, default_value=None,
  message_type=None, enum_type=None, containing_type=None,
  is_extension=True, extension_scope=None,
  serialized_options=None, file=DESCRIPTOR)
MESSAGE_FIELD_NUMBER = 1020
message = _descriptor.FieldDescriptor(
  name='message', full_name='scalapb.message', index=1,
  number=1020, type=11, cpp_type=10, label=1,
  has_default_value=False, default_value=None,
  message_type=None, enum_type=None, containing_type=None,
  is_extension=True, extension_scope=None,
  serialized_options=None, file=DESCRIPTOR)
FIELD_FIELD_NUMBER = 1020
field = _descriptor.FieldDescriptor(
  name='field', full_name='scalapb.field', index=2,
  number=1020, type=11, cpp_type=10, label=1,
  has_default_value=False, default_value=None,
  message_type=None, enum_type=None, containing_type=None,
  is_extension=True, extension_scope=None,
  serialized_options=None, file=DESCRIPTOR)


_SCALAPBOPTIONS = _descriptor.Descriptor(
  name='ScalaPbOptions',
  full_name='scalapb.ScalaPbOptions',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='package_name', full_name='scalapb.ScalaPbOptions.package_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='flat_package', full_name='scalapb.ScalaPbOptions.flat_package', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='import', full_name='scalapb.ScalaPbOptions.import', index=2,
      number=3, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=68,
  serialized_end=144,
)


_MESSAGEOPTIONS = _descriptor.Descriptor(
  name='MessageOptions',
  full_name='scalapb.MessageOptions',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='extends', full_name='scalapb.MessageOptions.extends', index=0,
      number=1, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=146,
  serialized_end=179,
)


_FIELDOPTIONS = _descriptor.Descriptor(
  name='FieldOptions',
  full_name='scalapb.FieldOptions',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='type', full_name='scalapb.FieldOptions.type', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=181,
  serialized_end=209,
)

DESCRIPTOR.message_types_by_name['ScalaPbOptions'] = _SCALAPBOPTIONS
DESCRIPTOR.message_types_by_name['MessageOptions'] = _MESSAGEOPTIONS
DESCRIPTOR.message_types_by_name['FieldOptions'] = _FIELDOPTIONS
DESCRIPTOR.extensions_by_name['options'] = options
DESCRIPTOR.extensions_by_name['message'] = message
DESCRIPTOR.extensions_by_name['field'] = field
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

ScalaPbOptions = _reflection.GeneratedProtocolMessageType('ScalaPbOptions', (_message.Message,), dict(
  DESCRIPTOR = _SCALAPBOPTIONS,
  __module__ = 'scalapb.scalapb_pb2'
  # @@protoc_insertion_point(class_scope:scalapb.ScalaPbOptions)
  ))
_sym_db.RegisterMessage(ScalaPbOptions)

MessageOptions = _reflection.GeneratedProtocolMessageType('MessageOptions', (_message.Message,), dict(
  DESCRIPTOR = _MESSAGEOPTIONS,
  __module__ = 'scalapb.scalapb_pb2'
  # @@protoc_insertion_point(class_scope:scalapb.MessageOptions)
  ))
_sym_db.RegisterMessage(MessageOptions)

FieldOptions = _reflection.GeneratedProtocolMessageType('FieldOptions', (_message.Message,), dict(
  DESCRIPTOR = _FIELDOPTIONS,
  __module__ = 'scalapb.scalapb_pb2'
  # @@protoc_insertion_point(class_scope:scalapb.FieldOptions)
  ))
_sym_db.RegisterMessage(FieldOptions)

options.message_type = _SCALAPBOPTIONS
google_dot_protobuf_dot_descriptor__pb2.FileOptions.RegisterExtension(options)
message.message_type = _MESSAGEOPTIONS
google_dot_protobuf_dot_descriptor__pb2.MessageOptions.RegisterExtension(message)
field.message_type = _FIELDOPTIONS
google_dot_protobuf_dot_descriptor__pb2.FieldOptions.RegisterExtension(field)

DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)



from __future__ import print_function

import inspect
import logging
import logging.config
import sys
import warnings

from segmind.tracking import fluent

# Logging format example:
# 2018/11/20 12:36:37 INFO cral.tracking.sagemaker: Creating new SageMaker endpoint # noqa: E501
LOGGING_LINE_FORMAT = '%(asctime)s %(levelname)s %(name)s: %(message)s'
LOGGING_DATETIME_FORMAT = '%Y/%m/%d %H:%M:%S'


def try_mlflow_log(fn, *args, **kwargs):
    """Catch exceptions and log a warning to avoid autolog throwing."""
    try:
        fn(*args, **kwargs)
    except Exception as e:  # pylint: disable=broad-except
        warnings.warn(
            'Logging to Refine server failed: ' + str(e), stacklevel=2)


def log_params_decorator(fn):

    def log_fn_args_as_params(*args, **kwargs):  # pylint: disable=W0102
        """Log parameters explicitly passed to a function.

        :param fn: function whose parameters are to be logged
        :param args: arguments explicitly passed into fn
        :param kwargs: kwargs explicitly passed into fn
        :return: None
        """
        # all_default_values has length n, corresponding to values of the
        # last n elements in all_param_names
        all_param_names, _, _, all_default_values = inspect.getargspec(fn)

        unnamed_args_length = len(args)

        _args_as_dict = dict(zip(all_param_names[:unnamed_args_length], args))

        num_args_without_default_value = len(all_param_names) - len(
            all_default_values)
        default_param_names = all_param_names[num_args_without_default_value:]
        _defaults_as_dict = dict(zip(default_param_names, all_default_values))
        _defaults_as_dict.update(**kwargs)

        _args_as_dict.update(**_defaults_as_dict)

        result = fn(*args, **kwargs)

        try_mlflow_log(fluent.log_params, _args_as_dict)

        return result

    return log_fn_args_as_params


def _configure_mlflow_loggers(root_module_name):
    logging.config.dictConfig({
        'version': 1,
        'disable_existing_loggers': False,
        'formatters': {
            'mlflow_formatter': {
                'format': LOGGING_LINE_FORMAT,
                'datefmt': LOGGING_DATETIME_FORMAT,
            },
        },
        'handlers': {
            'mlflow_handler': {
                'level': 'INFO',
                'formatter': 'mlflow_formatter',
                'class': 'logging.StreamHandler',
                'stream': sys.stderr,
            },
        },
        'loggers': {
            root_module_name: {
                'handlers': ['mlflow_handler'],
                'level': 'INFO',
                'propagate': False,
            },
        },
    })


def eprint(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)



from mlflow_lite.entities._mlflow_object import _MLflowObject
from mlflow_lite.protos.service_lite_pb2 import ArtifactTag as ProtoArtifactTag


class ArtifactTag(_MLflowObject):
    """ArtifactTag object."""

    def __init__(self, key, value):
        self._key = key
        self._value = value

    def __eq__(self, other):
        if type(other) is type(self):
            return self.__dict__ == other.__dict__
        return False

    @property
    def key(self):
        """String name of the tag."""
        return self._key

    @property
    def value(self):
        """String value of the tag."""
        return self._value

    def to_proto(self):
        artifact_tag = ProtoArtifactTag()
        artifact_tag.key = self.key
        artifact_tag.value = self.value
        return artifact_tag

    @classmethod
    def from_proto(cls, proto):
        return cls(proto.key, proto.value)



from segmind.store.artifact import artifact_repo
from segmind.store.tracking import abstract_store

__all__ = [
    # tracking server meta-data stores
    'abstract_store',

    # artifact repository stores
    'artifact_repo',
]



"""Internal module implementing the fluent API, allowing management of an
active MLflow run.

This module is exposed to users at the top-level :py:mod:`mlflow` module.

Attributes:
    NUM_RUNS_PER_PAGE_PANDAS (int): Description
    SEARCH_MAX_RESULTS_PANDAS (int): Description
"""

from __future__ import print_function

import atexit
import logging
import numpy as np
import os
import pandas as pd
import PIL.Image
import time
from google.protobuf.struct_pb2 import Struct
from tempfile import gettempdir

from segmind.entities import Metric, Param, Run, RunStatus, RunTag, ViewType
from segmind.entities.lifecycle_stage import LifecycleStage
from segmind.exceptions import MlflowException
from segmind.lite_extensions.client_utils import (_EXPERIMENT_ID_ENV_VAR,
                                                  _RUN_ID_ENV_VAR,
                                                  _get_experiment_id,
                                                  _runid_exists,
                                                  get_token)
from segmind.protos.service_lite_pb2 import RunTag as RunTagProto
from segmind.tracking.client import MlflowClient
from segmind.tracking.context import registry as context_registry
from segmind.utils import env
from segmind.utils.mlflow_tags import MLFLOW_PARENT_RUN_ID, MLFLOW_RUN_NAME
from segmind.utils.validation import _validate_run_id
from segmind.data.public import upload as upload_data

# _EXPERIMENT_ID_ENV_VAR = "MLFLOW_EXPERIMENT_ID"
# _EXPERIMENT_NAME_ENV_VAR = 'MLFLOW_EXPERIMENT_NAME'
# _RUN_ID_ENV_VAR = "MLFLOW_RUN_ID"
_active_run_stack = []
_active_experiment_id = None

SEARCH_MAX_RESULTS_PANDAS = 100000
NUM_RUNS_PER_PAGE_PANDAS = 10000

_logger = logging.getLogger(__name__)


def convert_to_imagefile(image):
    image_name = os.path.join(gettempdir(), 'temp_image.jpg')

    if isinstance(image, PIL.Image.Image):
        image.save(image_name)
        path = image_name
    elif isinstance(image, np.ndarray):
        image = PIL.Image.fromarray(image.astype(np.uint8))
        image.save(image_name)
        path = image_name
    else:
        path = image

    return path


def set_project(experiment_id):
    """Set given experiment as active experiment. If experiment does not exist,
    create an experiment with provided name.

    Args:
        experiment_id (str): id of experiment to be activated.

    Raises:
        MlflowException: Description
    """
    os.environ[_EXPERIMENT_ID_ENV_VAR] = experiment_id

    client = MlflowClient()

    experiment = client.get_experiment(experiment_id)
    if experiment_id is None:  # id can be 0
        print(f"INFO: '{experiment_id}' does not exist.")
        # experiment_id = client.create_experiment(experiment_name)
    elif experiment.lifecycle_stage == LifecycleStage.DELETED:
        raise MlflowException(
            "Cannot set a deleted experiment '%s' as the active experiment."
            ' You can restore the experiment, or permanently delete the '
            ' experiment to create a new one.' % experiment.name)
    global _active_experiment_id
    _active_experiment_id = experiment_id


def set_runid(run_id):
    """Summary.

    Args:
        run_id (TYPE): Description
    """
    os.environ[_RUN_ID_ENV_VAR] = run_id


class ActiveRun(Run):  # pylint: disable=W0223
    """Wrapper around :py:class:`segmind_track.entities.Run` to enable using
    Python ``with`` syntax."""

    def __init__(self, run):
        """Summary.

        Args:
            run (TYPE): Description
        """
        Run.__init__(self, run.info, run.data)

    def __enter__(self):
        """Summary.

        Returns:
            TYPE: Description
        """
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Summary.

        Args:
            exc_type (TYPE): Description
            exc_val (TYPE): Description
            exc_tb (TYPE): Description

        Returns:
            TYPE: Description
        """
        status = RunStatus.FINISHED if exc_type is None else RunStatus.FAILED
        end_run(RunStatus.to_string(status))
        return exc_type is None


def start_run(run_name=None, nested=False):
    """Start a new MLflow run, setting it as the active run under which metrics
    and parameters will be logged. The return value can be used as a context
    manager within a ``with`` block; otherwise, you must call ``end_run()`` to
    terminate the current run.

    If you pass a ``run_id`` or the ``MLFLOW_RUN_ID`` environment variable is
    set, ``start_run`` attempts to resume a run with the specified run ID and
    other parameters are ignored. ``run_id`` takes precedence over
    ``MLFLOW_RUN_ID``.

    MLflow sets a variety of default tags on the run, as defined in
    :ref:`MLflow system tags <system_tags>`.


    Args:
        run_id: If specified, get the run with the specified UUID and log
            parameters and metrics under that run. The run's end time is unset
            and its status is set to running, but the run's other attributes
            (``source_version``, ``source_type``, etc.) are not changed.
        experiment_id: ID of the experiment under which to create the current
            run (applicable only when ``run_id`` is not specified). If
            ``experiment_id`` argument is unspecified, will look for valid
            experiment in the following order: activated using
            ``set_project``, ``MLFLOW_EXPERIMENT_NAME`` environment
            variable, ``MLFLOW_EXPERIMENT_ID`` environment variable,
            or the default experiment as defined by the tracking server.
        run_name: Name of new run (stored as a ``segmind_track.runName`` tag).
            Used only when ``run_id`` is unspecified.
        nested: Controls whether run is nested in parent run. ``True`` creates
            a nest run.

    Returns:
        :py:class:`segmind_track.ActiveRun`: object that acts as a context
            manager wrappings the run's state.


    Raises:
        Exception: Description
        MlflowException: Description
    """
    global _active_run_stack
    # back compat for int experiment_id

    experiment_id = str(_get_experiment_id())

    if _runid_exists():
        existing_run_id = env.get_env(_RUN_ID_ENV_VAR)
    else:
        existing_run_id = None
    if len(_active_run_stack) > 0 and not nested:
        raise Exception(
            ('Run with UUID {} is already active. To start a new run, first ' +
             'end the current run with segmind_track.end_run().' +
             ' To start a nested run, call start_run with nested=True').format(
                 _active_run_stack[0].info.run_id))

    if existing_run_id is not None:
        _validate_run_id(existing_run_id)
        active_run_obj = MlflowClient().get_run(existing_run_id)
        # Check to see if experiment_id from environment matches experiment_id
        # from set_project()
        if (_active_experiment_id is not None and
                _active_experiment_id != active_run_obj.info.experiment_id):
            raise MlflowException(
                'Cannot start run with ID {} because active run ID '
                'does not match environment run ID. Make sure '
                '--experiment-name or --experiment-id matches experiment '
                'set with set_project(), or just use command-line '
                'arguments'.format(existing_run_id))
        # Check to see if current run isn't deleted
        if active_run_obj.info.lifecycle_stage == LifecycleStage.DELETED:
            raise MlflowException(
                'Cannot start run with ID {} because it is in the '
                'deleted state.'.format(existing_run_id))

    else:
        if len(_active_run_stack) > 0:
            parent_run_id = _active_run_stack[-1].info.run_id
        else:
            parent_run_id = None

        exp_id_for_run = experiment_id

        user_specified_tags = {}
        if parent_run_id is not None:
            user_specified_tags[MLFLOW_PARENT_RUN_ID] = parent_run_id
        if run_name is not None:
            user_specified_tags[MLFLOW_RUN_NAME] = run_name

        tags = context_registry.resolve_tags(user_specified_tags)

        active_run_obj = MlflowClient().create_run(
            experiment_id=exp_id_for_run, tags=tags)

    _active_run_stack.append(ActiveRun(active_run_obj))
    return _active_run_stack[-1]


def get_experiment_run(experiment_id=None, run_id=None, create_run_if_not_exists=False):
    experiment_obj = MlflowClient().get_experiment(experiment_id=experiment_id)
    os.environ[_EXPERIMENT_ID_ENV_VAR] = experiment_obj.experiment_id
    if not run_id:
        pass
    else:
        run_obj = MlflowClient().get_run(run_id=run_id)

    return experiment_obj, run_obj


def end_run(status=RunStatus.to_string(RunStatus.FINISHED)):
    """End an active MLflow run (if there is one).

    Args:
        status (TYPE, optional): Description
    """
    global _active_run_stack
    if len(_active_run_stack) > 0:
        MlflowClient().set_terminated(_active_run_stack[-1].info.run_id,
                                      status)
        # Clear out the global existing run environment variable as well.
        env.unset_variable(_RUN_ID_ENV_VAR)
        _active_run_stack.pop()


atexit.register(end_run)


def active_run():
    """Get the currently active ``Run``, or None if no such run exists.

    **Note**: You cannot access currently-active run attributes
    (parameters, metrics, etc.) through the run returned by
    ``segmind_track.active_run``. In order to access such attributes,
    use the :py:class:`segmind_track.tracking.MlflowClient` as follows:

    .. code-block:: py

        client = segmind_track.tracking.MlflowClient()
        data = client.get_run(segmind_track.active_run().info.run_id).data

    Returns:
        TYPE: Description
    """
    return _active_run_stack[-1] if len(_active_run_stack) > 0 else None


def get_run(run_id):
    """
    Fetch the run from backend store. The resulting
    :py:class:`Run <segmind_track.entities.Run>` contains a collection of run
    metadata -- :py:class:`RunInfo <segmind_track.entities.RunInfo>`, as well
    as a collection of run parameters, tags, and metrics --
    :py:class:`RunData <segmind_track.entities.RunData>`. In the case where
    multiple metrics with the same key are logged for the run, the
    :py:class:`RunData <segmind_track.entities.RunData>` contains
    the most recently logged value at the largest step for each metric.

    Args:
        run_id (TYPE): Unique identifier for the run.

    Returns:
        :py:class:`segmind_track.entities.Run`: object, if the run exists.
        Otherwise, raises an exception.
    """
    return MlflowClient().get_run(run_id)


def _get_or_start_run():
    """Summary.

    Returns:
        TYPE: Description
    """
    if len(_active_run_stack) > 0:
        return _active_run_stack[-1]

    return start_run()



def set_tag(key, value):
    """Set a tag under the current run. If no run is active, this method will
    create a new active run.

    Args:
        key (str): Tag name
        value (str): Tag value (string, but will be string-ified if not)
    """
    run_id = _get_or_start_run().info.run_id
    MlflowClient().set_tag(run_id, key, value)


def delete_tag(key):
    """Delete a tag from a run. This is irreversible. If no run is active, this
    method will create a new active run.

    Args:
        key (str): Name of the tag
    """
    run_id = _get_or_start_run().info.run_id
    MlflowClient().delete_tag(run_id, key)


def log_metric(key, value, step=None, epoch=None, tags={'sys_metric': 'no'}):
    """Log a metric under the current run. If no run is active, this method
    will create a new active run.

    Args:
        key (str): Metric name (string).
        value (float): Metric value (float). Note that some special values
                such as +/- Infinity may be replaced by other values depending
                on the store. For example, sFor example, the SQLAlchemy store
                replaces +/- Inf with max / min float values.
        step (int, optional): Metric step (int). Defaults to zero if
                            unspecified.
    """
    tags_arr = [RunTagProto(
        key=key,
        value=str(value)) for key, value in tags.items()]
    run_id = _get_or_start_run().info.run_id
    MlflowClient().log_metric(run_id, key, value, int(time.time() * 1000), step
                              or 0, epoch=epoch or 0, tags=tags_arr)


def log_param(key, value, tags={'sys_param': 'no'}):
    """Log a parameter under the current run. If no run is active, this method
    will create a new active run.

    Args:
        key (str: Parameter name (string)
        value (str): Parameter value (string, but will be string-ified if not)
    """
    tags_arr = [RunTagProto(
        key=key,
        value=str(value)) for key, value in tags.items()]
    run_id = _get_or_start_run().info.run_id
    MlflowClient().log_param(run_id, key, value, tags=tags_arr)


def log_params(params, tags={'sys_param': 'no'}):
    """Log a batch of params for the current run. If no run is active, this
    method will create a new active run.

    Args:
        params (dict): Dictionary of param_name: String -> value: (String, but
                    will be string-ified if not)
    """
    run_id = _get_or_start_run().info.run_id
    tags_arr = [RunTagProto(
        key=key,
        value=str(value)) for key, value in tags.items()]
    params_arr = [Param(
        key,
        str(value),
        tags=tags_arr) for key, value in params.items()]
    MlflowClient().log_batch(
        run_id=run_id, metrics=[], params=params_arr, tags=tags_arr)


def log_metrics(metrics, step=None, epoch=0, tags={'sys_metric': 'no'}):
    """Log multiple metrics for the current run. If no run is active, this
    method will create a new active run.

    Args:
        metrics (TYPE): Dictionary of metric_name: String -> value: Float.
                Note that some special values such as +/- Infinity may be
                replaced by other values depending on the store. For example,
                sql based store may replace +/- Inf with max / min float
                values.
        step (None, optional): A single integer step at which to log the
                specified Metrics. If unspecified, each metric is logged at
                step zero.
    """
    run_id = _get_or_start_run().info.run_id
    timestamp = int(time.time() * 1000)
    tags_arr = [RunTagProto(
        key=key,
        value=str(value)) for key, value in tags.items()]
    metrics_arr = [
        Metric(
            key,
            value,
            timestamp,
            step or 0,
            epoch=epoch or 0,
            tags=tags_arr)
        for key, value in metrics.items()
    ]
    MlflowClient().log_batch(
        run_id=run_id, metrics=metrics_arr, params=[], tags=tags_arr)


def log_batch(metrics={}, params={}, tags={}, step=None, epoch=None):
    """Summary.

    Args:
        metrics (dict, optional): Description
        params (dict, optional): Description
        tags (dict, optional): Description
        step (None, optional): Description
    """
    run_id = _get_or_start_run().info.run_id
    tags_arr = [RunTagProto(
        key=key,
        value=str(value)) for key, value in tags.items()]
    params_arr = [Param(key, str(value)) for key, value in params.items()]

    timestamp = int(time.time() * 1000)
    metrics_arr = [
        Metric(key, value, timestamp, step or 0, epoch or 0)
        for key, value in metrics.items()
    ]

    MlflowClient().log_batch(
        run_id=run_id, metrics=metrics_arr, params=params_arr, tags=tags_arr)


def log_artifact(key, path, step=None):
    """Log a local file or directory as an artifact of the currently active
    run. If no run is active, this method will create a new active run.

    Args:
        key (str): Name of the artifact to upload.
        path (str): Path of the artifact to upload.
        step (None, optional): integer indicating the step at which artifact
                            was generated
    """
    run = _get_or_start_run()
    run_id = run.info.run_id
    experiment_id = run.info.experiment_id
    MlflowClient().log_artifact_lite(
        run_id, experiment_id, key, path, step=step)


def log_image(key, image, tags={}, step=None):
    """logs an image artifact.

    Args:
        key (str): name of the table
        image (np.ndarray, PIL.Image): a numpy array of np.uint8 dtype or a
                                    PIL.Image object
        step (None, optional): integer indicating the step at which artifact
                            was generated
    """

    path = convert_to_imagefile(image)

    run = _get_or_start_run()
    run_id = run.info.run_id
    experiment_id = run.info.experiment_id
    MlflowClient().log_artifact_lite(
        run_id,
        experiment_id,
        key,
        path,
        artifact_type='image',
        step=step,
        tags=tags)


def log_table(key, table, step=None):
    """logs a pandas dataframe/csv file as a table artifact.

    Args:
        key (str): name of the table
        table (pd.Dataframe, str): A pandas dataframe or path to a csv file
        step (None, optional): integer indicating the step at which artifact
                            was generated
    Raises:
        MlflowException: Description
    """

    if not isinstance(table, (str, pd.DataFrame)):
        raise MlflowException(
            'table must be a pandas.DataFrame or a string to a csv file')

    path = table
    if isinstance(table, pd.DataFrame):
        path = os.path.join(gettempdir(), 'temp_table.csv')
        table.to_csv(path, index=False)

    run = _get_or_start_run()
    run_id = run.info.run_id
    experiment_id = run.info.experiment_id
    MlflowClient().log_artifact_lite(
        run_id, experiment_id, key, path, artifact_type='table', step=step)


# def log_bbox_prediction(key,
#                         image,
#                         bbox_pred,
#                         bbox_gt=None,
#                         bbox_type='pascal_voc',
#                         step=None):
#     """logs artifact for object detection.
#
#     Args:
#         key (str): name of the image
#         image (np.ndarray, PIL.Image): a numpy array of np.uint8 dtype or a
#                                     PIL.Image object
#         bbox_pred (np.ndarray, list): a list or a np.ndarray of dimension
#                                     (Nx4). All elemnts will be onverted to
#                                     int32.
#         bbox_gt (None, optional): a list or a np.ndarray of dimension (Nx4).
#                                     All elemnts will be onverted to int32
#         bbox_type (str, optional): can be one of 'yolo', 'pascal_voc', 'coco'
#                                 indicating the format of bbox and prediction.
#         step (None, optional): integer indicating the step at which artifact
#                             was generated.
#
#     Raises:
#         MlflowException: Description
#     """
#
#     if bbox_type not in ['pascal_voc', 'coco', 'yolo']:
#         raise MlflowException(
#             f'bbox_type should be one-of "pascal_voc, coco, yolo" not \
#             {bbox_type}')
#
#     path = convert_to_imagefile(image)
#
#     bbox_pred = np.array(bbox_pred)
#     assert isinstance(
#         bbox_pred, np.ndarray) and bbox_pred.ndim == 2 and bbox_pred.shape[
#             1] == 4, f'bbox_pred should be numpy of dimension (Nx4), got \
#         {bbox_pred.shape}'
#
#     if bbox_type == 'coco':
#         bbox_pred = coco_to_voc_bbox(bbox_pred)
#         if bbox_gt:
#             bbox_gt = coco_to_voc_bbox(bbox_gt)
#     elif bbox_type == 'yolo':
#         bbox_pred = yolo_to_voc_bbox(path, bbox_pred)
#         if bbox_gt:
#             bbox_gt = yolo_to_voc_bbox(path, bbox_gt)
#
#     if bbox_gt is None:
#         bbox_gt = np.array([])
#
#     run = _get_or_start_run()
#     run_id = run.info.run_id
#     experiment_id = run.info.experiment_id
#
#     prediction_struct = Struct()
#     prediction_struct.update({'bbox': bbox_pred.tolist()})
#
#     ground_truth_struct = Struct()
#     ground_truth_struct.update({'bbox': bbox_gt.tolist()})
#
#     MlflowClient().log_artifact_lite(
#         run_id,
#         experiment_id,
#         key,
#         path,
#         prediction=prediction_struct,
#         ground_truth=ground_truth_struct,
#         artifact_type='image',
#         step=step,
#         tags={'image_type': 'bbox_prediction'})


# def log_mask_prediction(key,
#                         image,
#                         pred_mask,
#                         bbox_pred=[],
#                         mask_gt=None,
#                         bbox_gt=None,
#                         bbox_type='pascal_voc',
#                         step=None):
#     """logs artifact for instance/semantic segmentation.
#
#     Args:
#         key (str): name of the image
#         image (np.ndarray, PIL.Image): a numpy array of np.uint8 dtype or a
#                                     PIL.Image object
#         pred_mask (np.ndarray, PIL.Image): a numpy array of np.uint8 dtype or
#                                     a PIL.Image object
#         bbox_pred (None, optional): Description
#         mask_gt (np.ndarray, PIL.Image): a numpy array of np.uint8 dtype or a
#                                     PIL.Image object
#         bbox_gt (None, optional): a list or a np.ndarray of dimension (Nx4).
#                                     All elemnts will be onverted to int32
#         bbox_type (str, optional): Description
#         step (None, optional): integer indicating the step at which artifact
#                                     was generated
#     """
#     if bbox_type not in ['pascal_voc', 'coco', 'yolo']:
#         raise MlflowException(
#             f'bbox_type should be one-of "pascal_voc, coco, yolo" not \
#             {bbox_type}')
#
#     path = convert_to_imagefile(image)
#     pred_mask_path = convert_to_imagefile(pred_mask)
#
#     bbox_pred = np.array(bbox_pred)
#     if bbox_pred.size > 0:
#         print(bbox_pred.size)
#         assert isinstance(
#             bbox_pred, np.ndarray) and bbox_pred.ndim == 2 and bbox_pred.shape[
#                 1] == 4, f'bbox_pred should be numpy of dimension (Nx4), got \
#             {bbox_pred.shape}'
#
#     if bbox_type == 'coco':
#         if bbox_pred.size > 0:
#             bbox_pred = coco_to_voc_bbox(bbox_pred)
#         if bbox_gt:
#             bbox_gt = coco_to_voc_bbox(bbox_gt)
#         else:
#             bbox_gt = np.array([])
#     else:
#         if bbox_pred.size > 0:
#             bbox_pred = yolo_to_voc_bbox(image, bbox_pred)
#         if bbox_gt:
#             bbox_gt = yolo_to_voc_bbox(image, bbox_gt)
#         else:
#             bbox_gt = np.array([])
#
#     run = _get_or_start_run()
#     run_id = run.info.run_id
#     experiment_id = run.info.experiment_id
#
#     prediction_struct = Struct()
#     prediction_struct.update({'bbox': bbox_pred.tolist()})
#
#     ground_truth_struct = Struct()
#     ground_truth_struct.update({'bbox': bbox_gt.tolist()})
#
#     MlflowClient().log_artifact_lite(
#         run_id,
#         experiment_id,
#         key,
#         path,
#         prediction=prediction_struct,
#         ground_truth=ground_truth_struct,
#         artifact_type='image',
#         step=step,
#         tags={'image_type': 'segmentation_mask'})
#
#     log_image(
#         key=key+'_mask',
#         image=pred_mask_path,
#         tags={'mask_parent': key},
#         step=step)


def set_tags(tags):
    """Log a batch of tags for the current run. If no run is active, this
    method will create a new active run.

    Args:
        tags (dict): Dictionary of tag_name: String -> value: (String, but
                    will be string-ified if not
    """
    run_id = _get_or_start_run().info.run_id
    tags_arr = [RunTag(key, str(value)) for key, value in tags.items()]
    MlflowClient().log_batch(
        run_id=run_id, metrics=[], params=[], tags=tags_arr)


def _record_logged_model(mlflow_model):
    """Summary.

    Args:
        mlflow_model (TYPE): Description
    """
    run_id = _get_or_start_run().info.run_id
    MlflowClient()._record_logged_model(run_id, mlflow_model)


def get_experiment(experiment_id):
    """Retrieve an experiment by experiment_id from the backend store.

    Args:
        experiment_id (TYPE): The experiment ID returned from
                            ``create_experiment``.

    Returns:
        TYPE: :py:class:`segmind_track.entities.Experiment`
    """
    return MlflowClient().get_experiment(experiment_id)


def get_experiment_by_name(name):
    """Retrieve an experiment by experiment name from the backend store.

    Args:
        name (TYPE): The experiment name.

    Returns:
        TYPE: :py:class:`segmind_track.entities.Experiment`
    """
    return MlflowClient().get_experiment_by_name(name)


def create_experiment(name, artifact_location=None):
    """Create an experiment.

    Args:
        name (TYPE): The experiment name. Must be unique.
        artifact_location (None, optional): The location to store run
                                        artifacts. If not provided,
                                        the server picks an appropriate
                                        default.

    Returns:
        int: Integer ID of the created experiment.
    """
    return MlflowClient().create_experiment(name, artifact_location)


def delete_experiment(experiment_id):
    """Delete an experiment from the backend store.

    Args:
        experiment_id (int): The experiment ID returned from
                            ``create_experiment``.
    """
    MlflowClient().delete_experiment(experiment_id)


def delete_run(run_id):
    """Deletes a run with the given ID.

    Args:
        run_id (TYPE): Unique identifier for the run to delete.
    """
    MlflowClient().delete_run(run_id)


def search_runs(experiment_ids=None,
                filter_string='',
                run_view_type=ViewType.ACTIVE_ONLY,
                max_results=SEARCH_MAX_RESULTS_PANDAS,
                order_by=None):
    """Get a pandas DataFrame of runs that fit the search criteria.

    Args:
        experiment_ids (None, optional): List of experiment IDs. None will
                                    default to the active experiment.
        filter_string (str, optional): Filter query string, defaults to
                                    searching all runs.
        run_view_type (TYPE, optional): one of enum values ``ACTIVE_ONLY``,
                             ``DELETED_ONLY``, or ``ALL`` runs defined
                            in :py:class:`segmind_track.entities.ViewType`.
        max_results (TYPE, optional): The maximum number of runs to put in the
                            dataframe. Default is 100,000 to avoid causing
                            out-of-memory issues on the user's machine.
        order_by (None, optional): List of columns to order by (e.g.,
                        "metrics.rmse"). The ``order_by`` column can contain
                        an optional ``DESC`` or ``ASC`` value. The default is
                        ``ASC``. The default ordering is to sort by
                        ``start_time DESC``, then ``run_id``.

    Returns:
        TYPE: A pandas.DataFrame of runs, where each metric, parameter, and
         tag are expanded into their own columns named metrics.*, params.*,
         and tags.* respectively. For runs that don't have a particular metric
        , parameter, or tag, their value will be (NumPy) Nan, None, or None
         respectively.
    """
    if not experiment_ids:
        experiment_ids = _get_experiment_id()
    runs = _get_paginated_runs(experiment_ids, filter_string, run_view_type,
                               max_results, order_by)
    info = {
        'run_id': [],
        'experiment_id': [],
        'status': [],
        'artifact_uri': [],
        'start_time': [],
        'end_time': []
    }
    params, metrics, tags = ({}, {}, {})
    PARAM_NULL, METRIC_NULL, TAG_NULL = (None, np.nan, None)
    for i, run in enumerate(runs):
        info['run_id'].append(run.info.run_id)
        info['experiment_id'].append(run.info.experiment_id)
        info['status'].append(run.info.status)
        info['artifact_uri'].append(run.info.artifact_uri)
        info['start_time'].append(
            pd.to_datetime(run.info.start_time, unit='ms', utc=True))
        info['end_time'].append(
            pd.to_datetime(run.info.end_time, unit='ms', utc=True))

        # Params
        param_keys = set(params.keys())
        for key in param_keys:
            if key in run.data.params:
                params[key].append(run.data.params[key])
            else:
                params[key].append(PARAM_NULL)
        new_params = set(run.data.params.keys()) - param_keys
        for p in new_params:
            params[p] = [PARAM_NULL
                         ] * i  # Fill in null values for all previous runs
            params[p].append(run.data.params[p])

        # Metrics
        metric_keys = set(metrics.keys())
        for key in metric_keys:
            if key in run.data.metrics:
                metrics[key].append(run.data.metrics[key])
            else:
                metrics[key].append(METRIC_NULL)
        new_metrics = set(run.data.metrics.keys()) - metric_keys
        for m in new_metrics:
            metrics[m] = [METRIC_NULL] * i
            metrics[m].append(run.data.metrics[m])

        # Tags
        tag_keys = set(tags.keys())
        for key in tag_keys:
            if key in run.data.tags:
                tags[key].append(run.data.tags[key])
            else:
                tags[key].append(TAG_NULL)
        new_tags = set(run.data.tags.keys()) - tag_keys
        for t in new_tags:
            tags[t] = [TAG_NULL] * i
            tags[t].append(run.data.tags[t])

    data = {}
    data.update(info)
    for key in metrics:
        data['metrics.' + key] = metrics[key]
    for key in params:
        data['params.' + key] = params[key]
    for key in tags:
        data['tags.' + key] = tags[key]
    return pd.DataFrame(data)


def _get_paginated_runs(experiment_ids, filter_string, run_view_type,
                        max_results, order_by):
    """Summary.

    Args:
        experiment_ids (TYPE): Description
        filter_string (TYPE): Description
        run_view_type (TYPE): Description
        max_results (TYPE): Description
        order_by (TYPE): Description

    Returns:
        TYPE: Description
    """
    all_runs = []
    next_page_token = None
    while (len(all_runs) < max_results):
        runs_to_get = max_results - len(all_runs)
        if runs_to_get < NUM_RUNS_PER_PAGE_PANDAS:
            runs = MlflowClient().search_runs(experiment_ids, filter_string,
                                              run_view_type, runs_to_get,
                                              order_by, next_page_token)
        else:
            runs = MlflowClient().search_runs(experiment_ids, filter_string,
                                              run_view_type,
                                              NUM_RUNS_PER_PAGE_PANDAS,
                                              order_by, next_page_token)
        all_runs.extend(runs)
        if hasattr(runs,
                   'token') and runs.token != '' and runs.token is not None:
            next_page_token = runs.token
        else:
            break
    return all_runs


def _get_experiment_id_from_env():
    """Summary.

    Returns:
        TYPE: Description
    """
    # experiment_name = env.get_env(_EXPERIMENT_NAME_ENV_VAR)
    # if experiment_name is not None:
    #     exp = MlflowClient().get_experiment_by_name(experiment_name)
    #     return exp.experiment_id if exp else None
    return env.get_env(_EXPERIMENT_ID_ENV_VAR)



"""The ``cral.tracking.entities`` module defines entities returned by the
MLflow `REST API <../rest-api.html>`_."""

from segmind.entities.artifact import Artifact
from segmind.entities.experiment import Experiment
from segmind.entities.experiment_tag import ExperimentTag
from segmind.entities.file_info import FileInfo
from segmind.entities.lifecycle_stage import LifecycleStage
from segmind.entities.metric import Metric
from segmind.entities.param import Param
from segmind.entities.run import Run
from segmind.entities.run_data import RunData
from segmind.entities.run_info import RunInfo
from segmind.entities.run_status import RunStatus
from segmind.entities.run_tag import RunTag
from segmind.entities.source_type import SourceType
from segmind.entities.view_type import ViewType

__all__ = [
    'Experiment', 'FileInfo', 'Metric', 'Param', 'Artifact', 'Run', 'RunData',
    'RunInfo', 'RunStatus', 'RunTag', 'ExperimentTag', 'SourceType',
    'ViewType', 'LifecycleStage'
]



"""An MLflow tracking server has two properties related to how data is stored:

*backend store* to record ML experiments, runs, parameters, metrics, etc., and
*artifact store* to store run artifacts like models, plots, images, etc.

Several constants are used by multiple backend store implementations.
"""

# Path to default location for backend when using local FileStore or ArtifactStore.
# Also used as default location for artifacts, when not provided, in non local file based backends
# (eg MySQL)

from segmind.lite_extensions.server_only.file_utils import _ROOT_DIR

DEFAULT_LOCAL_FILE_AND_ARTIFACT_PATH = _ROOT_DIR  #"./mlruns"
SEARCH_MAX_RESULTS_DEFAULT = 1000
SEARCH_MAX_RESULTS_THRESHOLD = 50000



import os
import posixpath
import tempfile
from abc import ABCMeta, abstractmethod

from segmind.exceptions import MlflowException
from segmind.protos.errorcodes_pb2 import (INVALID_PARAMETER_VALUE,
                                           RESOURCE_DOES_NOT_EXIST)
from segmind.utils.validation import bad_path_message, path_not_unique


class ArtifactRepository:
    """Abstract artifact repo that defines how to upload (log) and download
    potentially large artifacts from different storage backends."""

    __metaclass__ = ABCMeta

    def __init__(self, artifact_uri):
        self.artifact_uri = artifact_uri

    @abstractmethod
    def log_artifact(self, local_file, artifact_path=None):
        """Log a local file as an artifact, optionally taking an
        ``artifact_path`` to place it in within the run's artifacts. Run
        artifacts can be organized into directories, so you can place the
        artifact in a directory this way.

        Args:
            local_file: Path to artifact to log
            artifact_path: Directory within the run's artifact directory in
                        which to log the artifact.
        """

    @abstractmethod
    def log_artifacts(self, local_dir, artifact_path=None):
        """Log the files in the specified local directory as artifacts,
        optionally taking an ``artifact_path`` to place them in within the
        run's artifacts.

        Args:
        local_dir: Directory of local artifacts to log
        artifact_path: Directory within the run's artifact directory in which
                    to log the artifacts
        """

    @abstractmethod
    def list_artifacts(self, path):
        """Return all the artifacts for this run_id directly under path. If
        path is a file, returns an empty list. Will error if path is neither a
        file nor directory.

        :param path: Relative source path that contains desired artifacts

        :return: List of artifacts as FileInfo listed directly under path.
        """

    def _is_directory(self, artifact_path):
        listing = self.list_artifacts(artifact_path)
        return len(listing) > 0

    def download_artifacts(self, artifact_path, dst_path=None):
        """Download an artifact file or directory to a local directory if
        applicable, and return a local path for it. The caller is responsible
        for managing the lifecycle of the downloaded artifacts.

        Args:
        artifact_path: Relative source path to the desired artifacts.
        dst_path: Absolute path of the local filesystem destination directory
                to which to download the specified artifacts. This directory
                must already exist. If unspecified, the artifacts will either
                be downloaded to a new uniquely-named directory on the local
                filesystem or will be returned directly in the case of the
                LocalArtifactRepository.

        :return: Absolute path of the local filesystem location containing the
                desired artifacts.
        """

        if dst_path is None:
            dst_path = tempfile.mkdtemp()
        dst_path = os.path.abspath(dst_path)
        if not os.path.exists(dst_path):
            raise MlflowException(
                message=(
                    'The destination path for downloaded artifacts does not' +
                    f' exist! Destination path: {dst_path}'),
                error_code=RESOURCE_DOES_NOT_EXIST)
        elif not os.path.isdir(dst_path):
            raise MlflowException(
                message=(f'The destination path for downloaded artifacts must\
                be a directory! Destination path: {dst_path}'),
                error_code=INVALID_PARAMETER_VALUE)

        def download_file(fullpath):
            dirpath, _ = posixpath.split(fullpath)
            local_dir_path = os.path.join(dst_path, dirpath)
            local_file_path = os.path.join(dst_path, fullpath)
            if not os.path.exists(local_dir_path):
                os.makedirs(local_dir_path)
            self._download_file(
                remote_file_path=fullpath, local_path=local_file_path)
            return local_file_path

        def download_artifact_dir(dir_path):
            local_dir = os.path.join(dst_path, dir_path)
            dir_content = [
                file_info for file_info in self.list_artifacts(dir_path)
                if file_info.path != '.' and file_info.path != dir_path
            ]
            if not dir_content:  # empty dir
                if not os.path.exists(local_dir):
                    os.makedirs(local_dir)
            else:
                for file_info in dir_content:
                    if file_info.is_dir:
                        download_artifact_dir(dir_path=file_info.path)
                    else:
                        download_file(file_info.path)
            return local_dir

        if not os.path.exists(dst_path):
            raise MlflowException(
                message=(
                    'The destination path for downloaded artifacts does not' +
                    f' exist! Destination path: {dst_path}'),
                error_code=RESOURCE_DOES_NOT_EXIST)
        elif not os.path.isdir(dst_path):
            raise MlflowException(
                message=(f'The destination path for downloaded artifacts must\
                must be a directory! Destination path: {dst_path}'),
                error_code=INVALID_PARAMETER_VALUE)

        # Check if the artifacts points to a directory
        if self._is_directory(artifact_path):
            return download_artifact_dir(artifact_path)
        else:
            return download_file(artifact_path)

    @abstractmethod
    def _download_file(self, remote_file_path, local_path):
        """Download the file at the specified relative remote path and saves it
        at the specified local path.

        Args:
        remote_file_path: Source path to the remote file, relative to the root
                        directory of the artifact repository.
        local_path: The path to which to save the downloaded file.
        """


def verify_artifact_path(artifact_path):
    if artifact_path and path_not_unique(artifact_path):
        raise MlflowException(f"Invalid artifact path: '{artifact_path}'.\
         {bad_path_message(artifact_path)}")
